import mlprocessors as mlpr
import os
import time
import numpy as np
from os.path import join
from subprocess import Popen, PIPE
import shlex
import random
import string
import shutil
from spikeforest import mdaio
import spikeextractors as se
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
from mountaintools import client as mt
import json
import traceback
from .install_ironclust import install_ironclust


class IronClust(mlpr.Processor):
    NAME = 'IronClust'
    VERSION = '0.3.10'
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS', 'TEMPDIR']
    ADDITIONAL_FILES = ['*.m']
    CONTAINER = None

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    firings_out = mlpr.Output('Output firings file')
    channels = mlpr.IntegerListParameter(
        optional=True, default=[],
        description='List of channels to use.'
    )
    detect_sign = mlpr.IntegerParameter(
        optional=True, default=-1,
        description='Use -1, 0, or 1, depending on the sign of the spikes in the recording'
    )
    adjacency_radius = mlpr.FloatParameter(
        optional=True, default=50,
        description='Use -1 to include all channels in every neighborhood'
    )
    adjacency_radius_out = mlpr.FloatParameter(
        optional=True, default=75,
        description='Use -1 to include all channels in every neighborhood'
    )
    detect_threshold = mlpr.FloatParameter(
        optional=True, default=4.5,
        description='detection threshold'
    )
    prm_template_name = mlpr.StringParameter(
        optional=True, default='',
        description='.prm template file name'
    )
    freq_min = mlpr.FloatParameter(
        optional=True, default=300,
        description='Use 0 for no bandpass filtering'
    )
    freq_max = mlpr.FloatParameter(
        optional=True, default=6000,
        description='Use 0 for no bandpass filtering'
    )
    merge_thresh = mlpr.FloatParameter(
        optional=True, default=0.985,
        description='Threshold for automated merging'
    )
    pc_per_chan = mlpr.IntegerParameter(
        optional=True, default=2,
        description='Number of principal components per channel'
    )

    # added in version 0.2.4
    whiten = mlpr.BoolParameter(
        optional=True, default=False, description='Whether to do channel whitening as part of preprocessing')
    filter_type = mlpr.StringParameter(
        optional=True, default='bandpass', description='{none, bandpass, wiener, fftdiff, ndiff}')
    filter_detect_type = mlpr.StringParameter(
        optional=True, default='none', description='{none, bandpass, wiener, fftdiff, ndiff}')
    common_ref_type = mlpr.StringParameter(
        optional=True, default='none', description='{none, mean, median}')
    batch_sec_drift = mlpr.FloatParameter(
        optional=True, default=300, description='batch duration in seconds. clustering time duration')
    step_sec_drift = mlpr.FloatParameter(
        optional=True, default=20, description='compute anatomical similarity every n sec')
    knn = mlpr.IntegerParameter(
        optional=True, default=30, description='K nearest neighbors')
    min_count = mlpr.IntegerParameter(
        optional=True, default=30, description='Minimum cluster size')
    fGpu = mlpr.BoolParameter(
        optional=True, default=True, description='Use GPU if available')
    fft_thresh = mlpr.FloatParameter(
        optional=True, default=0, description='FFT-based noise peak threshold')
    nSites_whiten = mlpr.IntegerParameter(
        optional=True, default=32, description='Number of adjacent channels to whiten')
    feature_type = mlpr.StringParameter(
        optional=True, default='gpca', description='{gpca, pca, vpp, vmin, vminmax, cov, energy, xcov}')

    @staticmethod
    def install():
        print('Auto-installing ironclust.')
        return install_ironclust(commit='238d151b5407fda7bf2a779c1105638c1c9e8292')

    def run(self):
        ironclust_path = os.environ.get('IRONCLUST_PATH_DEV', None)
        if ironclust_path:
            print('Using ironclust from IRONCLUST_PATH_DEV directory: {}'.format(ironclust_path))
        else:
            try:
                ironclust_path = IronClust.install()
            except:
                traceback.print_exc()
                raise Exception('Problem installing ironclust. You can set the IRONCLUST_PATH_DEV to force to use a particular path.')
        print('Using ironclust from: {}'.format(ironclust_path))

        code = ''.join(random.choice(string.ascii_uppercase)
                       for x in range(10))
        tmpdir = os.environ.get('TEMPDIR', '/tmp') + '/ironclust-tmp-' + code

        try:
            recording = SFMdaRecordingExtractor(self.recording_dir)
            params = read_dataset_params(self.recording_dir)
            if len(self.channels) > 0:
                recording = se.SubRecordingExtractor(
                    parent_recording=recording, channel_ids=self.channels)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)

            all_params = dict()
            for param0 in self.PARAMETERS:
                all_params[param0.name] = getattr(self, param0.name)
            sorting = ironclust_helper(
                recording=recording,
                tmpdir=tmpdir,
                ironclust_path=ironclust_path,
                params=params,
                **all_params)
            SFMdaSortingExtractor.write_sorting(
                sorting=sorting, save_path=self.firings_out)
        except:
            if os.path.exists(tmpdir):
                if not getattr(self, '_keep_temp_files', False):
                    shutil.rmtree(tmpdir)
            raise
        if not getattr(self, '_keep_temp_files', False):
            pass
            # shutil.rmtree(tmpdir)


def ironclust_helper(
        *,
        recording,  # Recording object
        tmpdir,  # Temporary working directory
        params=dict(),
        ironclust_path,
        **kwargs):
    source_dir = os.path.dirname(os.path.realpath(__file__))

    dataset_dir = tmpdir + '/ironclust_dataset'
    # Generate three files in the dataset directory: raw.mda, geom.csv, params.json
    SFMdaRecordingExtractor.write_recording(
        recording=recording, save_path=dataset_dir, params=params, _preserve_dtype=True)

    samplerate = recording.get_sampling_frequency()

    print('Reading timeseries header...')
    HH = mdaio.readmda_header(dataset_dir + '/raw.mda')
    num_channels = HH.dims[0]
    num_timepoints = HH.dims[1]
    duration_minutes = num_timepoints / samplerate / 60
    print('Num. channels = {}, Num. timepoints = {}, duration = {} minutes'.format(
        num_channels, num_timepoints, duration_minutes))

    print('Creating argfile.txt...')
    txt = ''
    for key0, val0 in kwargs.items():
        txt += '{}={}\n'.format(key0, val0)
    txt += 'samplerate={}\n'.format(samplerate)
    if 'scale_factor' in params:
        txt += 'scale_factor={}\n'.format(params["scale_factor"])
    _write_text_file(dataset_dir + '/argfile.txt', txt)

    # new method
    print('Running ironclust in {tmpdir}...'.format(tmpdir=tmpdir))
    cmd = '''
        addpath('{source_dir}');
        addpath('{ironclust_path}', '{ironclust_path}/matlab', '{ironclust_path}/matlab/mdaio');
        try
            p_ironclust('{tmpdir}', '{dataset_dir}/raw.mda', '{dataset_dir}/geom.csv', '', '', '{tmpdir}/firings.mda', '{dataset_dir}/argfile.txt');
        catch
            fprintf('----------------------------------------');
            fprintf(lasterr());
            quit(1);
        end
        quit(0);
    '''
    cmd = cmd.format(ironclust_path=ironclust_path, tmpdir=tmpdir, dataset_dir=dataset_dir, source_dir=source_dir)

    matlab_cmd = mlpr.ShellScript(cmd, script_path=tmpdir + '/run_ironclust.m', keep_temp_files=True)
    matlab_cmd.write()

    shell_cmd = '''
        #!/bin/bash
        cd {tmpdir}
        matlab -nosplash -nodisplay -r run_ironclust
    '''.format(tmpdir=tmpdir)
    shell_cmd = mlpr.ShellScript(shell_cmd, script_path=tmpdir + '/run_ironclust.sh', keep_temp_files=True)
    shell_cmd.write(tmpdir + '/run_ironclust.sh')
    shell_cmd.start()

    retcode = shell_cmd.wait()

    if retcode != 0:
        raise Exception('ironclust returned a non-zero exit code')

    # parse output
    result_fname = tmpdir + '/firings.mda'
    if not os.path.exists(result_fname):
        raise Exception('Result file does not exist: ' + result_fname)

    firings = mdaio.readmda(result_fname)
    sorting = se.NumpySortingExtractor()
    sorting.set_times_labels(firings[1, :], firings[2, :])
    return sorting


def _read_text_file(fname):
    with open(fname) as f:
        return f.read()


def _write_text_file(fname, str):
    with open(fname, 'w') as f:
        f.write(str)


def read_dataset_params(dsdir):
    # ca = _load_required_modules()
    fname1 = dsdir + '/params.json'
    fname2 = mt.realizeFile(path=fname1)
    if not fname2:
        raise Exception('Unable to find file: ' + fname1)
    if not os.path.exists(fname2):
        raise Exception('Dataset parameter file does not exist: ' + fname2)
    with open(fname2) as f:
        return json.load(f)



from .spikeforest_views.currentstateview import CurrentStateView
from .analysis_views.analysissummaryview import AnalysisSummaryView
from .analysis_views.sorterdefinitionsview import SorterDefinitionsView
from .analysis_views.testview import TestView

import vdomr as vd
from mountaintools import client as mt
import json


def get_analysis_view_launchers(context):
    analysis_context = context

    launchers = []
    groups = []
    ret = dict(
        groups=groups,
        launchers=launchers
    )

    # General
    groups.append(dict(name='general', label=''))

    launchers.append(dict(
        group='general', name='analysis-summary', label='Analysis summary',
        view_class=AnalysisSummaryView,
        context=analysis_context, opts=dict(),
        enabled=True
    ))
    launchers.append(dict(
        group='general', name='sorter-definitions', label='Sorter definitions',
        view_class=SorterDefinitionsView,
        context=analysis_context, opts=dict(),
        enabled=True
    ))
    launchers.append(dict(
        group='general', name='test-view', label='Test view',
        view_class=TestView,
        context=analysis_context, opts=dict(),
        enabled=True
    ))
    launchers.append(dict(
        group='general', name='current-state', label='Current state',
        view_class=CurrentStateView,
        context=analysis_context, opts=dict(),
        enabled=True
    ))

    return ret



from .timeserieswidget import TimeseriesWidget, precomputeMultiscaleRecordings
from .templatewidget import TemplateWidget
from .featurespacewidget import FeatureSpaceWidget, FeatureSpaceWidgetPlotly
from .electrodegeometrywidget import ElectrodeGeometryWidget
from .unitwaveformswidget import UnitWaveformWidget, UnitWaveformsWidget
from .correlogramswidget import CorrelogramsWidget



import time
import os
from copy import deepcopy
from mountaintools import client as mt
import mlprocessors as mlpr
import multiprocessing
# from matplotlib import pyplot as plt
from .compute_units_info import ComputeUnitsInfo
from .computerecordinginfo import ComputeRecordingInfo
import mtlogging


@mtlogging.log()
def summarize_recordings(recordings, compute_resource=None, label=None, upload_to=None):
    print('')
    print('>>>>>> {}'.format(label or 'summarize recordings'))

    jobs_info = ComputeRecordingInfo.createJobs([
        dict(
            recording_dir=recording['directory'],
            channels=recording.get('channels', []),
            json_out={'ext': '.json', 'upload': True},
            _container='default',
            _label='Summarize recording: ' + recording.get('name', '')
        )
        for recording in recordings
    ])

    jobs_units_info = ComputeUnitsInfo.createJobs([
        dict(
            recording_dir=recording['directory'],
            firings=recording['directory'] + '/firings_true.mda',
            unit_ids=recording.get('units_true', None),
            channel_ids=recording.get('channels', None),
            json_out={'ext': '.json', 'upload': True},
            _container='default',
            _label='Compute units info for recording: ' + recording.get('name', '')
        )
        for recording in recordings
    ])

    # all_jobs=jobs_info+jobs_timeseries_plot+jobs_units_info
    all_jobs = jobs_info + jobs_units_info
    label = label or 'Summarize recordings'
    mlpr.executeBatch(jobs=all_jobs, label=label, num_workers=None, compute_resource=compute_resource)

    print('Gathering summarized recordings...')

    summarized_recordings = deepcopy(recordings)
    for ii in range(len(recordings)):
        summary = dict()

        result0 = jobs_info[ii].result
        summary['computed_info'] = mt.loadObject(path=result0.outputs['json_out'])
        summary['plots'] = dict()

        result0 = jobs_units_info[ii].result
        summary['true_units_info'] = mt.getSha1Url(path=result0.outputs['json_out'], basename='true_units_info.json')

        if upload_to:
            mt.saveFile(path=summary['true_units_info'], upload_to=upload_to)

        summarized_recordings[ii]['summary'] = summary

    return summarized_recordings


def _create_jobs_for_recording_old(recording):
    print('Creating jobs for recording: {}/{}'.format(recording.get('study', ''), recording.get('name', '')))
    dsdir = recording['directory']
    # raw_path=dsdir+'/raw.mda'
    firings_true_path = dsdir + '/firings_true.mda'
    # geom_path=dsdir+'/geom.csv'
    channels = recording.get('channels', None)
    units = recording.get('units_true', None)

    if not mt.findFile(path=firings_true_path):
        raise Exception('firings_true file not found: ' + firings_true_path)
    job_info = ComputeRecordingInfo.createJob(
        recording_dir=dsdir,
        channels=recording.get('channels', []),
        json_out={'ext': '.json', 'upload': True},
        _container='default'
    )
    job_info.addFilesToRealize([dsdir + '/raw.mda', dsdir + '/firings_true.mda'])
    # job=CreateTimeseriesPlot.createJob(
    #     recording_dir=dsdir,
    #     channels=recording.get('channels',[]),
    #     jpg_out={'ext':'.jpg','upload':True},
    #     _container='default'
    # )
    # jobs_timeseries_plot.append(job)
    job_units_info = ComputeUnitsInfo.createJob(
        recording_dir=dsdir,
        firings=dsdir + '/firings_true.mda',
        unit_ids=units,
        channel_ids=channels,
        json_out={'ext': '.json', 'upload': True},
        _container='default'
    )
    job_units_info.addFilesToRealize([dsdir + '/raw.mda', dsdir + '/firings_true.mda'])
    return (job_info, job_units_info)



from pathlib import Path

import mlprocessors as mlpr
import spikeextractors as se
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
import os, time, random, string, shutil, sys, shlex, json
from mountaintools import client as mt
from subprocess import Popen, PIPE, CalledProcessError, call


class Tridesclous(mlpr.Processor):
    """
    tridesclous is one of the more convenient, fast and elegant
    spike sorters.

    Installation instruction
        >>> pip install https://github.com/tridesclous/tridesclous/archive/master.zip

    More information on tridesclous at:
    * https://github.com/tridesclous/tridesclous
    * https://tridesclous.readthedocs.io

    """

    NAME = 'Tridesclous'
    VERSION = '0.1.1'  # wrapper VERSION
    ADDITIONAL_FILES = []
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS', 'TEMPDIR']
    CONTAINER = 'sha1://9fb4a9350492ee84c8ea5d8692434ecba3cf33da/2019-05-13/tridesclous.simg'
    LOCAL_MODULES = ['../../spikeforest']

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    firings_out = mlpr.Output('Output firings file')
    channels = mlpr.IntegerListParameter(optional=True, default=[],
                                         description='List of channels to use.')
    detect_sign = mlpr.FloatParameter(optional=True, default=-1, description='')
    detection_threshold = mlpr.FloatParameter(optional=True, default=5.5, description='')
    freq_min = mlpr.FloatParameter(optional=True, default=400, description='')
    freq_max = mlpr.FloatParameter(optional=True, default=5000, description='')
    waveforms_n_left = mlpr.IntegerParameter(description='', optional=True, default=-45)
    waveforms_n_right = mlpr.IntegerParameter(description='', optional=True, default=60)
    align_waveform = mlpr.BoolParameter(description='', optional=True, default=False)
    common_ref_removal = mlpr.BoolParameter(description='', optional=True, default=False)
    peak_span = mlpr.FloatParameter(optional=True, default=.0002, description='')
    alien_value_threshold = mlpr.FloatParameter(optional=True, default=100, description='')

    def run(self):
        import tridesclous as tdc

        tmpdir = Path(_get_tmpdir('tdc'))
        recording = SFMdaRecordingExtractor(self.recording_dir)

        params = {
            'fullchain_kargs': {
                'duration': 300.,
                'preprocessor': {
                    'highpass_freq': self.freq_min,
                    'lowpass_freq': self.freq_max,
                    'smooth_size': 0,
                    'chunksize': 1024,
                    'lostfront_chunksize': 128,
                    'signalpreprocessor_engine': 'numpy',
                    'common_ref_removal': self.common_ref_removal,
                },
                'peak_detector': {
                    'peakdetector_engine': 'numpy',
                    'peak_sign': '-',
                    'relative_threshold': self.detection_threshold,
                    'peak_span': self.peak_span,
                },
                'noise_snippet': {
                    'nb_snippet': 300,
                },
                'extract_waveforms': {
                    'n_left': self.waveforms_n_left,
                    'n_right': self.waveforms_n_right,
                    'mode': 'rand',
                    'nb_max': 20000,
                    'align_waveform': self.align_waveform,
                },
                'clean_waveforms': {
                    'alien_value_threshold': self.alien_value_threshold,
                },
            },
            'feat_method': 'peak_max',
            'feat_kargs': {},
            'clust_method': 'sawchaincut',
            'clust_kargs': {'kde_bandwith': 1.},
        }

        # save prb file:
        probe_file = tmpdir / 'probe.prb'
        se.save_probe_file(recording, probe_file, format='spyking_circus')

        # source file
        if isinstance(recording, se.BinDatRecordingExtractor) and recording._frame_first:
            # no need to copy
            raw_filename = recording._datfile
            dtype = recording._timeseries.dtype.str
            nb_chan = len(recording._channels)
            offset = recording._timeseries.offset
        else:
            # save binary file (chunk by hcunk) into a new file
            raw_filename = tmpdir / 'raw_signals.raw'
            n_chan = recording.get_num_channels()
            chunksize = 2**24 // n_chan
            se.write_binary_dat_format(recording, raw_filename, time_axis=0, dtype='float32', chunksize=chunksize)
            dtype = 'float32'
            offset = 0

        # initialize source and probe file
        tdc_dataio = tdc.DataIO(dirname=str(tmpdir))
        nb_chan = recording.get_num_channels()

        tdc_dataio.set_data_source(type='RawData', filenames=[str(raw_filename)],
                                   dtype=dtype, sample_rate=recording.get_sampling_frequency(),
                                   total_channel=nb_chan, offset=offset)
        tdc_dataio.set_probe_file(str(probe_file))

        try:
            sorting = tdc_helper(tmpdir=tmpdir, params=params, recording=recording)
            SFMdaSortingExtractor.write_sorting(
                sorting=sorting, save_path=self.firings_out)
        except:
            if os.path.exists(tmpdir):
                if not getattr(self, '_keep_temp_files', False):
                    shutil.rmtree(tmpdir)
            raise
        if not getattr(self, '_keep_temp_files', False):
            shutil.rmtree(tmpdir)


def tdc_helper(
        *,
        tmpdir,
        params,
        recording):

    import tridesclous as tdc

    # nb_chan = recording.get_num_channels()

    # check params and OpenCL when many channels
    use_sparse_template = False
    use_opencl_with_sparse = False
    # if nb_chan > 64:  # this limit depend on the platform of course
    #     if tdc.cltools.HAVE_PYOPENCL:
    #         # force opencl
    #         self.params['fullchain_kargs']['preprocessor']['signalpreprocessor_engine'] = 'opencl'
    #         use_sparse_template = True
    #         use_opencl_with_sparse = True
    #     else:
    #         print('OpenCL is not available processing will be slow, try install it')

    tdc_dataio = tdc.DataIO(dirname=str(tmpdir))
    # make catalogue
    chan_grps = list(tdc_dataio.channel_groups.keys())
    for chan_grp in chan_grps:
        cc = tdc.CatalogueConstructor(dataio=tdc_dataio, chan_grp=chan_grp)
        tdc.apply_all_catalogue_steps(cc, verbose=True, **params)
        cc.make_catalogue_for_peeler()

        # apply Peeler (template matching)
        initial_catalogue = tdc_dataio.load_catalogue(chan_grp=chan_grp)
        peeler = tdc.Peeler(tdc_dataio)
        peeler.change_params(catalogue=initial_catalogue,
                             use_sparse_template=use_sparse_template,
                             sparse_threshold_mad=1.5,
                             use_opencl_with_sparse=use_opencl_with_sparse,)
        peeler.run(duration=None, progressbar=False)

    sorting = se.TridesclousSortingExtractor(tmpdir)
    return sorting


# To be shared across sorters (2019.05.05)
def _get_tmpdir(sorter_name):
    code = ''.join(random.choice(string.ascii_uppercase) for x in range(10))
    tmpdir0 = os.environ.get('TEMPDIR', '/tmp')
    tmpdir = os.path.join(tmpdir0, '{}-tmp-{}'.format(sorter_name, code))
    # reset the output folder
    if os.path.exists(tmpdir):
        shutil.rmtree(str(tmpdir))
    else:
        os.makedirs(tmpdir)
    return tmpdir



from spikeextractors import RecordingExtractor
from spikeextractors import SortingExtractor

import json
import numpy as np
from .mdaio import DiskReadMda, readmda, writemda32, writemda64, writemda
import os
import mtlogging
import mlprocessors as mlpr


def _load_required_modules():
    try:
        from mountaintools import client as mt
    except ModuleNotFoundError:
        raise ModuleNotFoundError("To use the MdaExtractors, install mountainlab_pytools and kbucket: \n\n"
                                  "pip install mountainlab_pytools kbucket\n\n")
    return mt


class SFMdaRecordingExtractor(RecordingExtractor):
    def __init__(self, dataset_directory, *, download=True, raw_fname='raw.mda', params_fname='params.json'):
        ca = _load_required_modules()

        RecordingExtractor.__init__(self)
        self._dataset_directory = dataset_directory
        if '/' not in raw_fname:
            # relative path
            self._timeseries_path = dataset_directory + '/' + raw_fname
        else:
            # absolute path
            self._timeseries_path = raw_fname
        self._dataset_params = read_dataset_params(dataset_directory, params_fname)
        self._samplerate = self._dataset_params['samplerate'] * 1.0
        if download:
            path0 = ca.realizeFile(path=self._timeseries_path)
            if not path0:
                raise Exception('Unable to realize file: ' + self._timeseries_path)
            self._timeseries_path = path0

        geom0 = dataset_directory + '/geom.csv'
        self._geom_fname = ca.realizeFile(path=geom0)
        self._geom = np.genfromtxt(self._geom_fname, delimiter=',')

        timeseries_path_or_url = self._timeseries_path
        if not ca.isLocalPath(timeseries_path_or_url):
            a = ca.findFile(timeseries_path_or_url)
            if not a:
                raise Exception('Cannot find timeseries file: ' + timeseries_path_or_url)
            timeseries_path_or_url = a

        # if is_kbucket_url(timeseries0):
        #     download_needed = is_url(ca.findFile(path=timeseries0))
        # else:
        #     download_needed = is_url(timeseries0)
        # if download and download_needed:
        #     print('Downloading file: ' + timeseries0)
        #     self._timeseries_path = ca.realizeFile(path=timeseries0)
        #     print('Done.')
        # else:
        #     self._timeseries_path = ca.findFile(path=timeseries0)

        X = DiskReadMda(timeseries_path_or_url)
        if self._geom.shape[0] != X.N1():
            # raise Exception(
            #    'Incompatible dimensions between geom.csv and timeseries file {} <> {}'.format(self._geom.shape[0], X.N1()))
            print('WARNING: Incompatible dimensions between geom.csv and timeseries file {} <> {}'.format(self._geom.shape[0], X.N1()))
            self._geom = np.zeros((X.N1(), 2))

        self._num_channels = X.N1()
        self._num_timepoints = X.N2()
        for m in range(self._num_channels):
            self.set_channel_property(m, 'location', self._geom[m, :])

    def hash(self):
        from mountainclient import client as mt
        obj = dict(
            raw=mt.computeFileSha1(self._timeseries_path),
            geom=mt.computeFileSha1(self._geom_fname),
            params=self._dataset_params
        )
        return mt.sha1OfObject(obj)

    def recordingDirectory(self):
        return self._dataset_directory

    def get_channel_ids(self):
        return list(range(self._num_channels))

    def get_num_frames(self):
        return self._num_timepoints

    def get_sampling_frequency(self):
        return self._samplerate

    @mtlogging.log(name='SFMdaRecordingExtractor:get_traces')
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None):
        ca = _load_required_modules()
        if not ca.isLocalPath(self._timeseries_path):
            raise Exception('Cannot get traces -- timeseries file is not downloaded')
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.get_num_frames()
        if channel_ids is None:
            channel_ids = self.get_channel_ids()
        X = DiskReadMda(self._timeseries_path)
        recordings = X.readChunk(i1=0, i2=start_frame, N1=X.N1(), N2=end_frame - start_frame)
        recordings = recordings[channel_ids, :]
        return recordings

    @staticmethod
    def write_recording(recording, save_path, params=dict(), raw_fname='raw.mda', params_fname='params.json', 
            _preserve_dtype=False):
        # ca = _load_required_modules()
        channel_ids = recording.get_channel_ids()
        M = len(channel_ids)
        # N = recording.get_num_frames()
        raw = recording.get_traces()
        location0 = recording.get_channel_property(channel_ids[0], 'location')
        nd = len(location0)
        geom = np.zeros((M, nd))
        for ii in range(len(channel_ids)):
            location_ii = recording.get_channel_property(channel_ids[ii], 'location')
            geom[ii, :] = list(location_ii)
        if not os.path.isdir(save_path):
            os.mkdir(save_path)
        if _preserve_dtype:
            writemda(raw, save_path + '/' + raw_fname, dtype=raw.dtype)
        else:
            writemda32(raw, save_path + '/' + raw_fname)
        params["samplerate"] = recording.get_sampling_frequency()
        with open(save_path + '/' + params_fname, 'w') as f:
            json.dump(params, f)
        np.savetxt(save_path + '/geom.csv', geom, delimiter=',')


class SFMdaSortingExtractor(SortingExtractor):
    def __init__(self, firings_file):
        ca = _load_required_modules()

        SortingExtractor.__init__(self)
        if is_kbucket_url(firings_file):
            download_needed = is_url(ca.findFile(path=firings_file))
        else:
            download_needed = is_url(firings_file)
        if download_needed:
            print('Downloading file: ' + firings_file)
            self._firings_path = ca.realizeFile(path=firings_file)
            print('Done.')
        else:
            self._firings_path = ca.realizeFile(path=firings_file)
        if not self._firings_path:
            raise Exception('Unable to realize firings file: ' + firings_file)

        self._firings = readmda(self._firings_path)
        self._times = self._firings[1, :]
        self._labels = self._firings[2, :]
        self._unit_ids = np.unique(self._labels).astype(int)

    def get_unit_ids(self):
        return self._unit_ids

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        inds = np.where((self._labels == unit_id) & (start_frame <= self._times) & (self._times < end_frame))
        return np.rint(self._times[inds]).astype(int)

    def hash(self):
        from mountaintools import client as mt
        return mt.computeFileSha1(self._firings_path)

    @staticmethod
    def write_sorting(sorting, save_path):
        # ca = _load_required_modules()
        unit_ids = sorting.get_unit_ids()
        # if len(unit_ids) > 0:
        #     K = np.max(unit_ids)
        # else:
        #     K = 0
        times_list = []
        labels_list = []
        for i in range(len(unit_ids)):
            unit = unit_ids[i]
            times = sorting.get_unit_spike_train(unit_id=unit)
            times_list.append(times)
            labels_list.append(np.ones(times.shape) * unit)
        all_times = _concatenate(times_list)
        all_labels = _concatenate(labels_list)
        sort_inds = np.argsort(all_times)
        all_times = all_times[sort_inds]
        all_labels = all_labels[sort_inds]
        L = len(all_times)
        firings = np.zeros((3, L))
        firings[1, :] = all_times
        firings[2, :] = all_labels
        writemda64(firings, save_path)


def _concatenate(list):
    if len(list) == 0:
        return np.array([])
    return np.concatenate(list)


def is_kbucket_url(path):
    path = path or ''
    return path.startswith('kbucket://') or path.startswith('sha1://') or path.startswith('sha1dir://')


def is_url(path):
    path = path or ''
    return path.startswith('http://') or path.startswith('https://') or path.startswith(
        'kbucket://') or path.startswith('sha1://') or path.startswith('sha1dir://')


def read_dataset_params(dsdir, params_fname):
    ca = _load_required_modules()

    fname1 = dsdir + '/' + params_fname
    fname2 = ca.realizeFile(path=fname1)
    if not fname2:
        raise Exception('Unable to find file: ' + fname1)
    if not os.path.exists(fname2):
        raise Exception('Dataset parameter file does not exist: ' + fname2)
    with open(fname2) as f:
        return json.load(f)



import vdomr as vd
from .viewcontainer import ViewContainer
from .forestviewcontrolpanel import ForestViewControlPanel
import uuid
import multiprocessing
import traceback
import sys
import time
from mountaintools import client as mt

# MEDIUM TODO move tabs between north/south containers
# MEDIUM TODO cross-correlograms widget


class ForestViewMainWindow(vd.Component):
    def __init__(self, context):
        vd.Component.__init__(self)

        self._context = context
        self._control_panel = ForestViewControlPanel(self._context)
        self._view_container_north = ViewContainer()
        self._view_container_south = ViewContainer()
        self._view_containers = [self._view_container_north, self._view_container_south]
        self._control_panel.onLaunchView(self._trigger_launch_view)

        self._current_view_container = self._view_container_north
        self._view_container_north.onClick(self._on_click_north)
        self._view_container_south.onClick(self._on_click_south)

        # style0 = dict(border='solid 1px gray')
        self._container_CP = Container(self._control_panel, scroll=True)
        self._container_VCN = Container(self._view_container_north)
        self._container_VCS = Container(self._view_container_south)
        self._container_main = Container(self._container_CP, self._container_VCN, self._container_VCS, position_mode='relative')

        self._highlight_view_containers()

        self._size = (1200, 800)

        vd.devel.loadBootstrap()

        self._update_sizes()

    def setSize(self, size):
        self._size = size
        self._update_sizes()

    def size(self):
        return self._size

    def _update_sizes(self):
        width = self._size[0]
        # width1 = int(min(300, width*0.3))
        width1 = 320
        width2 = width - width1 - 30
        height = self._size[1]
        height1 = int(height / 2) - 5
        height2 = height - height1 - 30

        self._container_main.setSize((width, height))
        self._container_main.setPosition((0, 0))

        self._container_CP.setSize((width1, height - 20))
        self._container_CP.setPosition((10, 10))

        self._container_VCN.setSize((width2, height1))
        self._container_VCN.setPosition((width1 + 20, 10))

        self._container_VCS.setSize((width2, height2))
        self._container_VCS.setPosition((width1 + 20, height1 + 20))

        self._view_container_north.setSize((width2, height1))
        self._view_container_south.setSize((width2, height2))

    def _highlight_view_containers(self):
        for VC in [self._view_container_north, self._view_container_south]:
            VC.setHighlight(VC == self._current_view_container)

    def _on_click_north(self):
        self._current_view_container = self._view_container_north
        self._highlight_view_containers()

    def _on_click_south(self):
        self._current_view_container = self._view_container_south
        self._highlight_view_containers()

    def _trigger_launch_view(self, view_launcher):
        if not view_launcher.get('always_open_new', False):
            for VC in self._view_containers:
                v = VC.findView(name=view_launcher.get('name', ''))
                if v:
                    VC.setCurrentView(v)
                    return
        frame = ViewFrame(view_launcher=view_launcher)
        self._current_view_container.addView(frame, name=view_launcher.get('name', ''))
        frame.initialize()

    def context(self):
        return self._context

    def render(self):
        return self._container_main


class TextComponent(vd.Component):
    def __init__(self):
        vd.Component.__init__(self)
        self._text = ''

    def setText(self, txt):
        if self._text == txt:
            return
        self._text = txt
        self.refresh()

    def render(self):
        return vd.span(self._text)


class TitleBar(vd.Component):
    def __init__(self):
        vd.Component.__init__(self)
        self._title = 'TitleBar'
        self._height = 20

    def setTitle(self, title0):
        if title0 == self._title:
            return
        self._title = title0
        self.refresh()

    def height(self):
        return self._height

    def render(self):
        return vd.div(self._title, style={'height': '{}px'.format(self._height), 'font-size': '14px'})


class ViewFrame(vd.Component):
    def __init__(self, *, view_launcher):
        vd.Component.__init__(self)
        self._view_launcher = view_launcher
        self._connection_to_prepare = None
        self._prepare_log_text = ''
        self._prepare_log_text_view = TextComponent()
        self._title_bar = TitleBar()
        self._view = None
        self._size = (100, 100)
        self._init_process = None
        self.updateTitle()

    def setSize(self, size):
        self._size = size
        self._update_view_size()

    def size(self):
        return self._size

    def tabLabel(self):
        if self._view:
            return self._view.tabLabel()
        else:
            return self._view_launcher['label'] + '...'

    def updateTitle(self):
        if self._view:
            if hasattr(self._view, 'title'):
                title0 = self._view.title()
            else:
                title0 = ''
        else:
            title0 = 'Preparing: {} ...'.format(self._view_launcher['label'])
        self._title_bar.setTitle(title0)

    def initialize(self):
        view_launcher = self._view_launcher
        context = view_launcher['context']
        opts = view_launcher['opts']
        view_class = view_launcher['view_class']
        if hasattr(view_class, 'prepareView'):
            self._connection_to_prepare, connection_to_parent = multiprocessing.Pipe()
            self._init_process = multiprocessing.Process(
                target=_prepare_in_worker,
                args=(view_class, context, opts, connection_to_parent)
            )
            self._init_process.start()

            self._check_prepare_count = 0
            vd.set_timeout(self._check_prepare, 0.1)
        else:
            if hasattr(context, 'initialize'):
                context.initialize()
            self._view = view_class(context=context, opts=opts)
            self._update_view_size()
            self.updateTitle()
        self.refresh()

    def cleanup(self):
        if self._init_process:
            print('# terminating init process')
            self._init_process.terminate()
        if self._view:
            if hasattr(self._view, 'cleanup'):
                (getattr(self._view, 'cleanup'))()

    def render(self):
        if self._view:
            X = self._view
        else:
            X = vd.components.ScrollArea(
                vd.div(
                    vd.h3('Preparing...'),
                    vd.pre(self._prepare_log_text_view)
                ),
                height=self._size[1] - self._title_bar.height()
            )
        return vd.div(self._title_bar, X)

    def _update_view_size(self):
        size = self._size
        if self._view:
            self._view.setSize((size[0], size[1] - self._title_bar.height()))

    def _check_prepare(self):
        if not self._view:
            if self._connection_to_prepare.poll():
                msg = self._connection_to_prepare.recv()
                if msg['name'] == 'log':
                    self._prepare_log_text = self._prepare_log_text + msg['text']
                    self._prepare_log_text_view.setText(self._prepare_log_text)
                elif msg['name'] == 'result':
                    self._on_prepare_completed(msg['result'])
                    return
            self._check_prepare_count = self._check_prepare_count + 1
            if self._check_prepare_count < 3:
                timeout = 0.2
            elif self._check_prepare_count < 5:
                timeout = 0.5
            elif self._check_prepare_count < 10:
                timeout = 1
            else:
                timeout = 5
            vd.set_timeout(self._check_prepare, timeout)

    def _on_prepare_completed(self, result):
        self._init_process = None
        view_launcher = self._view_launcher
        context = view_launcher['context']
        opts = view_launcher['opts']
        view_class = view_launcher['view_class']
        if hasattr(context, 'initialize'):
            context.initialize()
        self._view = view_class(context=context, opts=opts, prepare_result=result)
        self._update_view_size()
        self.refresh()


class StdoutSender():
    def __init__(self, connection):
        self._connection = connection
        self._handler = _StdoutHandler(connection)

    def __enter__(self):
        self._old_stdout = sys.stdout
        self._old_stderr = sys.stderr
        self._handler.setOtherStdout(self._old_stdout)
        sys.stdout = self._handler
        sys.stderr = self._handler
        return dict()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._handler.send()
        sys.stdout = self._old_stdout
        sys.stderr = self._old_stderr


class _StdoutHandler(object):
    def __init__(self, connection):
        self._connection = connection
        self._text = ''
        self._timer = time.time()
        self._other_stdout = None

    def write(self, data):
        if self._other_stdout:
            self._other_stdout.write(data)
        self._text = self._text + str(data)
        elapsed = time.time() - self._timer
        if elapsed > 5:
            self.send()
            self._timer = time.time()

    def flush(self):
        if self._other_stdout:
            self._other_stdout.flush()

    def setOtherStdout(self, other_stdout):
        self._other_stdout = other_stdout

    def send(self):
        if self._text:
            self._connection.send(dict(name="log", text=self._text))
            self._text = ''


def _prepare_in_worker(view_class, context, opts, connection_to_parent):
    # mt.setDownloadFromConfig(download_from_config)
    with StdoutSender(connection=connection_to_parent):
        try:
            print('***** Preparing...')
            result0 = view_class.prepareView(context=context, opts=opts)
        except:
            traceback.print_exc()
            raise
    connection_to_parent.send(dict(
        name='result',
        result=result0
    ))


class Container(vd.Component):
    def __init__(self, *args, position=(0, 0), size=(0, 0), position_mode='absolute', style=dict(), scroll=False):
        vd.Component.__init__(self)
        self._elmt_id = 'Container-' + str(uuid.uuid4())
        self._children = list(args)
        self._position = position
        self._size = size
        self._position_mode = position_mode
        self._style = style
        self._scroll = scroll

    def setSize(self, size):
        js = """
    document.getElementById('{elmt_id}').style.width='{width}px';
    document.getElementById('{elmt_id}').style.height='{height}px';
    """
        js = js.replace('{elmt_id}', self._elmt_id)
        js = js.replace('{width}', str(size[0]))
        js = js.replace('{height}', str(size[1]))
        self.executeJavascript(js)

    def setPosition(self, position):
        js = """
    document.getElementById('{elmt_id}').style.left='{left}px';
    document.getElementById('{elmt_id}').style.top='{top}px';
    """
        js = js.replace('{elmt_id}', self._elmt_id)
        js = js.replace('{left}', str(position[0]))
        js = js.replace('{top}', str(position[1]))
        self.executeJavascript(js)

    def render(self):
        style = self._style
        style['position'] = self._position_mode
        style['width'] = '{}px'.format(self._size[0])
        style['height'] = '{}px'.format(self._size[1])
        style['left'] = '{}px'.format(self._position[0])
        style['top'] = '{}px'.format(self._position[1])
        if self._scroll:
            style['overflow'] = 'auto'
        else:
            style['overflow'] = 'hidden'
        ret = vd.div(
            self._children,
            style=style,
            id=self._elmt_id
        )
        # if self._scroll:
        #    ret = vd.components.ScrollArea(ret, height=self._size[1])
        return ret



from spikeforest import SFMdaRecordingExtractor
import numpy as np


def real(name='franklab_tetrode', download=True):
    if name == 'franklab_tetrode':
        dsdir = 'kbucket://b5ecdf1474c5/datasets/neuron_paper/franklab_tetrode'
        IX = SFMdaRecordingExtractor(dataset_directory=dsdir, download=download)
        return (IX, None)
    else:
        raise Exception('Unrecognized name for real dataset: ' + name)



import sys
import time


class _StdoutHandler(object):
    def __init__(self, connection):
        self._connection = connection
        self._text = ''
        self._timer = time.time()
        self._other_stdout = None

    def write(self, data):
        if self._other_stdout:
            self._other_stdout.write(data)
        self._text = self._text + str(data)
        elapsed = time.time() - self._timer
        if elapsed > 5:
            self.send()
            self._timer = time.time()

    def flush(self):
        if self._other_stdout:
            self._other_stdout.flush()

    def setOtherStdout(self, other_stdout):
        self._other_stdout = other_stdout

    def send(self):
        if self._text:
            self._connection.send(dict(name="log", text=self._text))
            self._text = ''


class StdoutSender():
    def __init__(self, connection):
        self._connection = connection
        self._handler = _StdoutHandler(connection)

    def __enter__(self):
        self._old_stdout = sys.stdout
        self._old_stderr = sys.stderr
        self._handler.setOtherStdout(self._old_stdout)
        sys.stdout = self._handler
        sys.stderr = self._handler
        return dict()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._handler.send()
        sys.stdout = self._old_stdout
        sys.stderr = self._old_stderr



from .summarize_recordings import summarize_recordings
from .sort_recordings import sort_recordings
from .sort_recordings import multi_sort_recordings
from .summarize_sortings import summarize_sortings
from .compare_sortings_with_truth import compare_sortings_with_truth, GenSortingComparisonTable
from .aggregate_sorting_results import aggregate_sorting_results
from .compute_units_info import compute_units_info, ComputeUnitsInfo
from .computerecordinginfo import ComputeRecordingInfo
from .bandpass_filter import bandpass_filter
from .whiten import whiten
from .sortingcomparison import SortingComparison
from .sort_recordings import find_sorter_processor_and_container
# from .sort_recordings import IronClust
# from .sort_recordings import KiloSort



from copy import deepcopy
from mountaintools import client as mt
from mountaintools import MountainClient
from .analysis_view_launchers import get_analysis_view_launchers

local_client = MountainClient()


class AnalysisContext():
    def __init__(self, obj):
        self._object = obj
        self._o = self._object  # for convenience
        self._any_state_change_handlers = []
        self._signal_handlers = dict()

        print('******** FORESTVIEW: Done initializing study context')
        self._state = dict(
            current_sorter=None
        )

    def initialize(self):
        pass

    def analysisName(self):
        return self._o.get('analysis_name', None)

    def outputPath(self):
        return self._o.get('output', None)

    def recordingGroups(self):
        return self._o.get('recordings', [])

    def sorterKeys(self):
        return self._o.get('sorter_keys', [])

    def downloadFrom(self):
        return self._o.get('download_from', [])

    def jobTimeout(self):
        return self._o.get('job_timeout', None)

    def computeResourceKeys(self):
        return self._o.get('compute_resources', dict()).keys()

    def sorterDefinitionKeys(self):
        return self._o.get('sorters', dict()).keys()

    def sorterDefinition(self, key):
        return self._o.get('sorters', dict()).get(key)

    def viewLaunchers(self):
        return get_analysis_view_launchers(self)

    def onAnyStateChanged(self, handler):
        for key in self._state.keys():
            self._register_state_change_handler(key, handler)
        self._any_state_change_handlers.append(handler)

    def stateObject(self):
        ret = deepcopy(self._state)
        return ret

    def _set_state_value(self, name, val):
        if self._state[name] == val:
            return
        self._state[name] = deepcopy(val)
        self._emit('state-changed-' + name)

    def _get_state_value(self, name):
        return deepcopy(self._state[name])

    def _register_state_change_handler(self, name, handler):
        self._register_signal_handler('state-changed-' + name, handler)

    def _register_signal_handler(self, signal_name, handler):
        if signal_name not in self._signal_handlers:
            self._signal_handlers[signal_name] = []
        self._signal_handlers[signal_name].append(handler)

    def _emit(self, signal_name):
        if signal_name in self._signal_handlers:
            for handler in self._signal_handlers[signal_name]:
                handler()

    def _trigger_any_state_change_handlers(self):
        for handler in self._any_state_change_handlers:
            handler()



import mlprocessors as mlpr
import os
import random
import string
import shutil
import sys
import shlex
import json
from mountaintools import client as mt


class HerdingSpikes2(mlpr.Processor):
    """
    HerdingSpikes2 wrapper for SpikeForest framework
    Calls hs2 wrapper in SpikeInterface/SpikeToolkit

    See: HILGEN, Gerrit, et al. Unsupervised spike sorting for large-scale,
    high-density multielectrode arrays. Cell reports, 2017, 18.10: 2521-2532.
    https://github.com/mhhennig/hs2
    """

    NAME = 'HS2'
    VERSION = '0.2.3'  # wrapper VERSION
    ADDITIONAL_FILES = []
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS', 'TEMPDIR']
    # CONTAINER = 'sha1://d140fc9b43b98a5f70a538970bc037b5b35fefd8/2019-05-08/herdingspikes2.simg'
    # CONTAINER = 'sha1://b3209735e078abe212dfa508e267786382d28473/2019-05-30/herdingspikes2.simg'
    CONTAINER = 'sha1://57b5bf971d44c5333a07cae1a8c188df6eb0e9a1/2019-06-05/herdingspikes2.simg'
    LOCAL_MODULES = ['../../spikeforest', '../../spikeforest_common']

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    firings_out = mlpr.Output('Output firings file')

    def run(self):
        from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
        from spikeforest_common import autoScaleRecordingToNoiseLevel
        import spiketoolkit as st

        clustering_n_jobs = os.environ.get('NUM_WORKERS', None)
        if clustering_n_jobs is not None:
            clustering_n_jobs = int(clustering_n_jobs)

        code = ''.join(random.choice(string.ascii_uppercase) for x in range(10))
        tmpdir = os.environ.get('TEMPDIR', '/tmp') + '/hs2-tmp-' + code

        try:
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)

            recording = SFMdaRecordingExtractor(self.recording_dir)
            # print('Auto scaling via normalize_by_quantile...')
            # recording = st.preprocessing.normalize_by_quantile(recording=recording, scale=200.0)
            # recording = autoScaleRecordingToNoiseLevel(recording, noise_level=32)

            print('Running HerdingspikesSorter...')
            os.environ['HS2_PROBE_PATH'] = tmpdir
            st_sorter = st.sorters.HerdingspikesSorter(
                recording=recording,
                output_folder=tmpdir + '/hs2_sorting_output'
            )
            print('Using builtin bandpass and normalisation')
            hs2_par = st_sorter.default_params()
            hs2_par['filter'] = True
            hs2_par['pre_scale'] = True
            st_sorter.set_params(**hs2_par)
            if clustering_n_jobs is not None:
                st_sorter.set_params(clustering_n_jobs=clustering_n_jobs)
            st_sorter.run()
            sorting = st_sorter.get_result()

            SFMdaSortingExtractor.write_sorting(
                sorting=sorting, save_path=self.firings_out)
        except:
            if os.path.exists(tmpdir):
                if not getattr(self, '_keep_temp_files', False):
                    shutil.rmtree(tmpdir)
            raise

        if not getattr(self, '_keep_temp_files', False):
            shutil.rmtree(tmpdir)



import sfdata as sf
import mtlogging


@mtlogging.log()
def aggregate_sorting_results(studies, recordings, sorting_results):
    SF = sf.SFData()
    SF.loadStudies(studies=studies)
    SF.loadRecordings2(recordings=recordings)
    SF.loadSortingResults(sorting_results=sorting_results)

    aggregated_sorting_results = dict(
        recording_sorting_results=[],
        study_sorting_results=[]
    )
    for study_name in SF.studyNames():
        print('study: ' + study_name)
        S = SF.study(study_name)
        if len(S.recordingNames()) > 0:
            first_recording = S.recording(S.recordingNames()[0])
            sorter_names = first_recording.sortingResultNames()

            for srname in sorter_names:
                print('sorter: ' + srname)

                study_results0 = dict(
                    recording_indices=[],
                    true_unit_ids=[],
                    true_unit_snrs=[],
                    true_unit_firing_rates=[],
                    num_matches=[],
                    num_false_positives=[],
                    num_false_negatives=[]
                )

                comparisons_all_exist = True
                some_comparison_exists = False
                for recording_index, rname in enumerate(S.recordingNames()):
                    rec = S.recording(rname)
                    SR = rec.sortingResult(srname)
                    comparison = SR.comparisonWithTruth(format='json')
                    if comparison is None:
                        comparisons_all_exist = False
                    else:
                        some_comparison_exists = True
                if not comparisons_all_exist:
                    print('WARNING: Comparisons do not all exist for sorter sorter: {}'.format(srname))
                if some_comparison_exists:
                    for recording_index, rname in enumerate(S.recordingNames()):
                        rec = S.recording(rname)
                        true_units_info = rec.trueUnitsInfo(format='json')
                        true_units_info_by_id = dict()
                        for true_unit in true_units_info:
                            true_units_info_by_id[true_unit['unit_id']] = true_unit
                        SR = rec.sortingResult(srname)
                        comparison = SR.comparisonWithTruth(format='json')
                        if comparison is None:
                            # put in a zero comparison
                            comparison = dict()
                            for true_unit in true_units_info:
                                comparison[true_unit['unit_id']] = dict(
                                    unit_id=true_unit['unit_id'],
                                    num_false_positives=0,
                                    num_false_negatives=true_unit['num_events'],
                                    num_matches=0
                                )
                        if comparison is not None:
                            recording_results0 = dict(
                                true_unit_ids=[],
                                true_unit_snrs=[],
                                true_unit_firing_rates=[],
                                num_matches=[],
                                num_false_positives=[],
                                num_false_negatives=[]
                            )

                            ok = True
                            for key0 in comparison.keys():
                                unit = comparison[key0]
                                # best_unit = unit['best_unit']
                                unit_id = unit['unit_id']
                                true_unit = true_units_info_by_id[unit_id]

                                recording_results0['true_unit_ids'].append(unit_id)
                                recording_results0['true_unit_snrs'].append(round(true_unit['snr'], 3))
                                recording_results0['true_unit_firing_rates'].append(
                                    round(true_unit['firing_rate'], 3))
                                if 'num_false_positives' in unit:
                                    recording_results0['num_matches'].append(unit['num_matches'])
                                    recording_results0['num_false_positives'].append(unit['num_false_positives'])
                                    recording_results0['num_false_negatives'].append(unit['num_false_negatives'])
                                else:
                                    ok = False
                                    error = 'missing field: num_false_positives'
                                    break
                            if ok:
                                recording_sorting_result = dict(
                                    study=study_name,
                                    recording=rname,
                                    sorter=srname,
                                    true_unit_ids=recording_results0['true_unit_ids'],
                                    true_unit_snrs=recording_results0['true_unit_snrs'],
                                    true_unit_firing_rates=recording_results0['true_unit_firing_rates'],
                                    num_matches=recording_results0['num_matches'],
                                    num_false_positives=recording_results0['num_false_positives'],
                                    num_false_negatives=recording_results0['num_false_negatives']
                                )
                                aggregated_sorting_results['recording_sorting_results'].append(recording_sorting_result)
                            else:
                                print('Warning: ' + error)

                            study_results0['recording_indices'].extend([recording_index] * len(recording_results0['true_unit_ids']))
                            study_results0['true_unit_ids'].extend(recording_results0['true_unit_ids'])
                            study_results0['true_unit_snrs'].extend(recording_results0['true_unit_snrs'])
                            study_results0['true_unit_firing_rates'].extend(recording_results0['true_unit_firing_rates'])
                            study_results0['num_matches'].extend(recording_results0['num_matches'])
                            study_results0['num_false_positives'].extend(recording_results0['num_false_positives'])
                            study_results0['num_false_negatives'].extend(recording_results0['num_false_negatives'])

                    study_sorting_result = dict(
                        study=study_name,
                        sorter=srname,
                        true_unit_recording_indices=study_results0['recording_indices'],
                        true_unit_ids=study_results0['true_unit_ids'],
                        true_unit_snrs=study_results0['true_unit_snrs'],
                        true_unit_firing_rates=study_results0['true_unit_firing_rates'],
                        num_matches=study_results0['num_matches'],
                        num_false_positives=study_results0['num_false_positives'],
                        num_false_negatives=study_results0['num_false_negatives']
                    )
                    aggregated_sorting_results['study_sorting_results'].append(study_sorting_result)
                else:
                    print('WARNING: Skipping aggregation for sorter because no comparisons exist: {}'.format(srname))

    return aggregated_sorting_results



import vdomr as vd
import time
import multiprocessing
import sys
from .stdoutsender import StdoutSender
import mtlogging
import numpy as np
import spikeforest_analysis as sa
from spikeforest import EfficientAccessRecordingExtractor
import json
import mlprocessors as mlpr
import traceback
from mountaintools import client as mt
from .tablewidget import TableWidget


class UnitsTableView(vd.Component):
    def __init__(self, context, opts=None, prepare_result=None):
        vd.Component.__init__(self)
        self._sorting_context = context
        self._recording_context = context.recordingContext()
        self._size = (100, 100)
        self._unit_table_widget = UnitTableWidget(
            context=self._sorting_context,
            units_info=prepare_result['units_info']
        )
        self._unit_table_widget.setSize(self._size)

    @staticmethod
    def prepareView(context, opts):
        sorting_context = context
        recording_context = context.recordingContext()
        try:
            recording_context.initialize()
            sorting_context.initialize()
            print('***** Preparing efficient access recording extractor...')
            earx = EfficientAccessRecordingExtractor(recording=recording_context.recordingExtractor())
            print('***** computing units info...')
            info0 = mt.loadObject(path=ComputeUnitsInfo.execute(recording=earx, sorting=sorting_context.sortingExtractor(), json_out=True).outputs['json_out'])
            print('*****')
        except:
            traceback.print_exc()
            raise
        return dict(
            units_info=info0
        )

    def setSize(self, size):
        self._size = size
        if self._unit_table_widget:
            self._unit_table_widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Units table'

    def render(self):
        if self._unit_table_widget:
            return vd.div(
                self._unit_table_widget
            )
        else:
            return vd.div(
                vd.h3('Initializing......'),
                style=dict(overflow='auto')
            )


class UnitTableWidget(vd.Component):
    def __init__(self, *, context, units_info):
        vd.Component.__init__(self)
        self._sorting_context = context
        self._recording_context = context.recordingContext()
        self._size = (100, 100)
        self._units_info = units_info
        self._update_table()
        self._sorting_context.onCurrentUnitIdChanged(self._on_context_selection_changed)
        self._sorting_context.onSelectedUnitIdsChanged(self._on_context_selection_changed)

    def setSize(self, size):
        self._size = size
        self._update_table()

    def _update_table(self):
        self._table_widget = TableWidget(
            columns=[
                dict(label='Unit ID', name='unit_id'),
                dict(label='SNR', name='snr'),
                dict(label='Peak channel', name='peak_channel'),
                dict(label='Num. events', name='num_events'),
                dict(label='Firing rate', name='firing_rate')
            ],
            records=self._units_info,
            height=self._size[1]
        )
        self._table_widget.onSelectionChanged(self._on_widget_selection_changed)
        self._on_context_selection_changed()
        self.refresh()

    def _on_widget_selection_changed(self):
        current_row_index = self._table_widget.currentRowIndex()
        if current_row_index is not None:
            unit_id = self._units_info[current_row_index]['unit_id']
            self._sorting_context.setCurrentUnitId(unit_id)
            if unit_id is not None:
                self._sorting_context.setSelectedUnitIds([unit_id])
            else:
                self._sorting_context.setSelectedUnitIds([])
        else:
            self._sorting_context.setCurrentUnitId(None)

    def _on_context_selection_changed(self):
        unit_id = self._sorting_context.currentUnitId()
        unit_ids = [info['unit_id'] for info in self._units_info]
        if unit_id is not None:
            try:
                index0 = unit_ids.index(unit_id)
            except:
                index0 = None
        else:
            index0 = None
        self._table_widget.setCurrentRowIndex(index0)

    def render(self):
        return self._table_widget


class ComputeUnitsInfo(mlpr.Processor):
    NAME = 'ComputeUnitsInfo'
    VERSION = '0.1.3'
    recording = mlpr.Input()
    sorting = mlpr.Input()
    json_out = mlpr.Output()

    def run(self):
        info0 = sa.compute_units_info(recording=self.recording, sorting=self.sorting)
        with open(self.json_out, 'w') as f:
            json.dump(info0, f)



from .widgets.timeserieswidget import TimeseriesWidget
from .widgets.featurespacewidget import FeatureSpaceWidget
from .widgets.unitwaveformswidget import UnitWaveformsWidget
from .widgets.sortingaccuracywidget import SortingAccuracyWidget
from .widgets.crosscorrelogramswidget import CrossCorrelogramsWidget
from .widgets.electrodegeometrywidget import ElectrodeGeometryWidget

from .tables.sortingcomparisontable import SortingComparisonTable

from .validation.sortingcomparison import SortingComparison

from . import example_datasets
from . import lazyfilters
from . import devel



from spikeextractors import SortingExtractor
import numpy as np


class NeuroscopeSortingExtractor(SortingExtractor):
    """
    Extracts sorting information from pair of .res and .clu files . The .res is a text file with
    a sorted list of all spiketimes from all units displayed in sample (integer '%i') units.
    The .clu file is a file with one more row than the .res with the first row corresponding to
    the total number of unit ids and the rest of the rows indicating which unit id the corresponding
    entry in the .res file refers to.

    Parameters
    ----------
    resfile : str
        Path to the .res text file.
    clufile : str
        Path to the .clu text file.
    """

    def __init__(self, resfile, clufile):
        SortingExtractor.__init__(self)
        res = np.loadtxt(resfile, dtype=np.int64, usecols=0, ndmin=1)
        clu = np.loadtxt(clufile, dtype=np.int64, usecols=0, ndmin=1)
        if len(res) > 0:
            n_clu = clu[0]
            clu = np.delete(clu, 0)
            self._spiketrains = []
            self._unit_ids = list(x + 1 for x in range(n_clu))
            for s_id in self._unit_ids:
                self._spiketrains.append(res[(clu == s_id).nonzero()])
        else:
            self._spiketrains = []
            self._unit_ids = []

    def get_unit_ids(self):
        return list(self._unit_ids)

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        times = self._spiketrains[self.get_unit_ids().index(unit_id)]
        inds = np.where((start_frame <= times) & (times < end_frame))
        return times[inds]

    @staticmethod
    def write_sorting(sorting, save_path):
        save_res = "{}.res".format(save_path)
        save_clu = "{}.clu".format(save_path)

        unit_ids = sorting.get_unit_ids()
        spiketrains = [sorting.get_unit_spike_train(u) for u in unit_ids]
        res = np.concatenate(spiketrains).ravel()
        clu = np.concatenate([np.repeat(i + 1, len(st)) for i, st in enumerate(spiketrains)]).ravel()
        res_sort = np.argsort(res)
        res = res[res_sort]
        clu = clu[res_sort]
        clu = np.insert(clu, 0, len(unit_ids))

        np.savetxt(save_res, res, fmt='%i')
        np.savetxt(save_clu, clu, fmt='%i')



import mlprocessors as mlpr
import os
import time
import numpy as np
from os.path import join
from subprocess import Popen, PIPE
import shlex
import random
import string
import shutil
import spikeextractors as se
from .tools import saveProbeFile
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
from .spykingcircussortingextractor import SpykingCircusSortingExtractor


class SpykingCircus(mlpr.Processor):
    NAME = 'SpykingCircus'
    VERSION = '0.2.3'
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS']
    ADDITIONAL_FILES = ['*.params']
    CONTAINER = 'sha1://8958530b960522d529163344af2faa09ea805716/2019-05-06/spyking_circus.simg'
    LOCAL_MODULES = ['../../spikeforest']

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    channels = mlpr.IntegerListParameter(
        description='List of channels to use.', optional=True, default=[])
    firings_out = mlpr.Output('Output firings file')

    detect_sign = mlpr.IntegerParameter(description='-1, 1, or 0')
    adjacency_radius = mlpr.FloatParameter(
        optional=True, default=100, description='Channel neighborhood adjacency radius corresponding to geom file')
    spike_thresh = mlpr.FloatParameter(
        optional=True, default=6, description='Threshold for detection')
    template_width_ms = mlpr.FloatParameter(
        optional=True, default=3, description='Spyking circus parameter')
    filter = mlpr.BoolParameter(optional=True, default=True)
    whitening_max_elts = mlpr.IntegerParameter(
        optional=True, default=1000, description='I believe it relates to subsampling and affects compute time')
    clustering_max_elts = mlpr.IntegerParameter(
        optional=True, default=10000, description='I believe it relates to subsampling and affects compute time')

    def run(self):
        tmpdir = _get_tmpdir('spiking-circus')
        num_workers = int(os.environ.get('NUM_WORKERS', 1))

        try:
            recording = SFMdaRecordingExtractor(self.recording_dir)
            if len(self.channels) > 0:
                recording = se.SubRecordingExtractor(
                    parent_recording=recording, channel_ids=self.channels)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)
            sorting = spyking_circus(
                recording=recording,
                output_folder=tmpdir,
                probe_file=None,
                file_name=None,
                detect_sign=self.detect_sign,
                adjacency_radius=self.adjacency_radius,
                spike_thresh=self.spike_thresh,
                template_width_ms=self.template_width_ms,
                filter=self.filter,
                merge_spikes=True,
                n_cores=num_workers,
                electrode_dimensions=None,
                whitening_max_elts=self.whitening_max_elts,
                clustering_max_elts=self.clustering_max_elts,
            )
            SFMdaSortingExtractor.write_sorting(
                sorting=sorting, save_path=self.firings_out)
        except:
            if not getattr(self, '_keep_temp_files', False):
                if os.path.exists(tmpdir):
                    shutil.rmtree(tmpdir)
            raise
        if not getattr(self, '_keep_temp_files', False):
            shutil.rmtree(tmpdir)


def spyking_circus(
    recording,
    output_folder=None,  # Temporary working directory
    probe_file=None,
    file_name=None,
    detect_sign=-1,  # -1 - 1 - 0
    adjacency_radius=100,  # Channel neighborhood adjacency radius corresponding to geom file
    spike_thresh=6,  # Threshold for detection
    template_width_ms=3,  # Spyking circus parameter
    filter=True,
    merge_spikes=True,
    n_cores=None,
    electrode_dimensions=None,
    whitening_max_elts=1000,  # I believe it relates to subsampling and affects compute time
    # I believe it relates to subsampling and affects compute time
    clustering_max_elts=10000
):
    try:
        import circus  # pylint: disable=import-error
    except ModuleNotFoundError:
        raise ModuleNotFoundError("\nTo use Spyking-Circus, install spyking-circus: \n\n"
                                  "\npip install spyking-circus"
                                  "\nfor ubuntu install openmpi: "
                                  "\nsudo apt install libopenmpi-dev"
                                  "\nMore information on Spyking-Circus at: "
                                  "\nhttps://spyking-circus.readthedocs.io/en/latest/")
    source_dir = os.path.dirname(os.path.realpath(__file__))

    if output_folder is None:
        output_folder = 'spyking_circus'
    else:
        output_folder = join(output_folder, 'spyking_circus')
    output_folder = os.path.abspath(output_folder)

    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)

    # save prb file:
    if probe_file is None:
        saveProbeFile(recording, join(output_folder, 'probe.prb'), format='spyking_circus', radius=adjacency_radius,
                      dimensions=electrode_dimensions)
        probe_file = join(output_folder, 'probe.prb')
    # save binary file
    if file_name is None:
        file_name = 'recording'
    elif file_name.endswith('.npy'):
        file_name = file_name[file_name.find('.npy')]
    raw = recording.get_traces()
    np.save(join(output_folder, file_name),
            raw.astype(raw.dtype, order='F'))

    if detect_sign < 0:
        detect_sign = 'negative'
    elif detect_sign > 0:
        detect_sign = 'positive'
    else:
        detect_sign = 'both'

    # set up spykingcircus config file
    with open(join(source_dir, 'config_default.params'), 'r') as f:
        circus_config = f.read()
    if merge_spikes:
        auto = 1e-5
    else:
        auto = 0
    circus_config = circus_config.format(
        float(recording.get_sampling_frequency()
              ), probe_file, template_width_ms, spike_thresh, detect_sign, filter,
        whitening_max_elts, clustering_max_elts, auto
    )
    params_file = join(output_folder, file_name + '.params')
    with open(params_file, 'w') as f:
        f.write(circus_config)

    # with open(params_file) as ff:
    #    print('CIRCUS CONFIG:')
    #    print(ff.read())

    print('Running spyking circus...')
    t_start_proc = time.time()
    if n_cores is None:
        n_cores = np.maximum(1, int(os.cpu_count() / 2))

    output_folder_cmd = output_folder

    num_cores_str = ''
    if int(n_cores) > 1:
        num_cores_str = '-c {}'.format(n_cores)
    cmd = 'spyking-circus {} {} '.format(
        join(output_folder_cmd, file_name + '.npy'), num_cores_str)

    # I think the merging step requires a gui and some user interaction. TODO: inquire about this
    # cmd_merge = 'spyking-circus {} -m merging {} '.format(join(output_folder_cmd, file_name+'.npy'), num_cores_str)
    # cmd_convert = 'spyking-circus {} -m converting'.format(join(output_folder, file_name+'.npy'))

    retcode = run_command_and_print_output(cmd)
    if retcode != 0:
        raise Exception('Spyking circus returned a non-zero exit code')

    # retcode = run_command_and_print_output(cmd_merge)
    # if retcode != 0:
    #    raise Exception('Spyking circus merging returned a non-zero exit code')
    processing_time = time.time() - t_start_proc
    print('Elapsed time: ', processing_time)
    sorting = SpykingCircusSortingExtractor(join(output_folder, file_name))

    return sorting


# To be shared across sorters (2019.05.05)
def _get_tmpdir(sorter_name):
    code = ''.join(random.choice(string.ascii_uppercase) for x in range(10))
    tmpdir0 = os.environ.get('TEMPDIR', '/tmp')
    tmpdir = os.path.join(tmpdir0, '{}-tmp-{}'.format(sorter_name, code))
    # reset the output folder
    if os.path.exists(tmpdir):
        shutil.rmtree(str(tmpdir))
    else:
        os.makedirs(tmpdir)
    return tmpdir


def run_command_and_print_output(command):
    print('RUNNING: ' + command)
    with Popen(shlex.split(command), stdout=PIPE, stderr=PIPE) as process:
        while True:
            output_stdout = process.stdout.readline()
            output_stderr = process.stderr.readline()
            if (not output_stdout) and (not output_stderr) and (process.poll() is not None):
                break
            if output_stdout:
                print(output_stdout.decode())
            if output_stderr:
                print(output_stderr.decode())
        rc = process.poll()
        return rc



import mlprocessors as mlpr
import json

import spikeextractors as si
from .sfmdaextractors import SFMdaRecordingExtractor, SFMdaSortingExtractor

#_CONTAINER = 'sha1://5627c39b9bd729fc011cbfce6e8a7c37f8bcbc6b/spikeforest_basic.simg'
_CONTAINER = 'sha1://0944f052e22de0f186bb6c5cb2814a71f118f2d1/spikeforest_basic.simg' #MAY26JJJ


# A MountainLab processor for generating the summary info for a recording
class ComputeRecordingInfo(mlpr.Processor):
    NAME = 'ComputeRecordingInfo'
    VERSION = '0.1.1'
    CONTAINER = _CONTAINER

    recording_dir = mlpr.Input(directory=True, description='Recording directory')
    channels = mlpr.IntegerListParameter(description='List of channels to use.', optional=True, default=[])
    json_out = mlpr.Output('Info in .json file')

    def run(self):
        ret = {}
        recording = SFMdaRecordingExtractor(dataset_directory=self.recording_dir, download=True)
        if len(self.channels) > 0:
            recording = si.SubRecordingExtractor(parent_recording=recording, channel_ids=self.channels)
        ret['samplerate'] = recording.get_sampling_frequency()
        ret['num_channels'] = len(recording.get_channel_ids())
        ret['duration_sec'] = recording.get_num_frames() / ret['samplerate']
        write_json_file(self.json_out, ret)


def write_json_file(fname, obj):
    with open(fname, 'w') as f:
        json.dump(obj, f)



import numpy as np
import spikeextractors as se
from scipy.optimize import linear_sum_assignment


class SortingComparison():
    def __init__(self, sorting1, sorting2, sorting1_name=None, sorting2_name=None, delta_tp=10, minimum_accuracy=0.5,
                 count=False, verbose=False):
        self._sorting1 = sorting1
        self._sorting2 = sorting2
        self.sorting1_name = sorting1_name
        self.sorting2_name = sorting2_name
        self._delta_tp = delta_tp
        self._min_accuracy = minimum_accuracy
        if verbose:
            print("Matching...")
        self._do_matching()

        self._counts = None
        if count:
            if verbose:
                print("Counting...")
            self._do_counting(verbose=False)

    def getSorting1(self):
        return self._sorting1

    def getSorting2(self):
        return self._sorting2

    def getLabels1(self, unit_id):
        if unit_id in self._sorting1.get_unit_ids():
            return self._labels_st1[unit_id]
        else:
            raise Exception("Unit_id is not a valid unit")

    def getLabels2(self, unit_id):
        if unit_id in self._sorting1.get_unit_ids():
            return self._labels_st1[unit_id]
        else:
            raise Exception("Unit_id is not a valid unit")

    def getMappedSorting1(self):
        return MappedSortingExtractor(self._sorting2, self._unit_map12)

    def getMappedSorting2(self):
        return MappedSortingExtractor(self._sorting1, self._unit_map21)

    def getMatchingEventCount(self, unit1, unit2):
        if (unit1 is not None) and (unit2 is not None):
            if unit1 != -1:
                a = self._matching_event_counts_12[unit1]
                if unit2 in a:
                    return a[unit2]
                else:
                    return 0
            else:
                return 0
        else:
            raise Exception('getMatchingEventCount: unit1 and unit2 must not be None.')

    def _compute_agreement_score(self, num_matches, num1, num2):
        denom = num1 + num2 - num_matches
        if denom == 0:
            return 0
        return num_matches / denom

    def _compute_safe_frac(self, numer, denom):
        if denom == 0:
            return 0
        return float(numer) / denom

    def getBestUnitMatch1(self, unit1):
        if unit1 in self._best_match_units_12:
            return self._best_match_units_12[unit1]
        else:
            return None

    def getBestUnitMatch2(self, unit2):
        if unit2 in self._best_match_units_21:
            return self._best_match_units_21[unit2]
        else:
            return None

    def getMatchingUnitList1(self, unit1):
        a = self._matching_event_counts_12[unit1]
        return list(a.keys())

    def getMatchingUnitList2(self, unit2):
        a = self._matching_event_counts_21[unit2]
        return list(a.keys())

    def getAgreementFraction(self, unit1=None, unit2=None):
        if (unit1 is not None) and (unit2 is None):
            if unit1 != -1:
                unit2 = self.getBestUnitMatch1(unit1)
                if unit2 is None or unit2 == -1:
                    return 0
            else:
                return 0
        if (unit1 is None) and (unit2 is not None):
            if unit1 != -1 and unit2 != -1:
                unit1 = self.getBestUnitMatch2(unit2)
                if unit1 is None or unit1 == -1:
                    return 0
            else:
                return 0
        if (unit1 is None) and (unit2 is None):
            raise Exception('getAgreementFraction: at least one of unit1 and unit2 must not be None.')

        if unit1 != -1 and unit2 != -1:
            a = self._matching_event_counts_12[unit1]
            if unit2 not in a:
                return 0
        else:
            return 0
        return self._compute_agreement_score(a[unit2], self._event_counts_1[unit1], self._event_counts_2[unit2])

    def getFalsePositiveFraction(self, unit1, unit2=None):
        if unit1 is None:
            raise Exception('getFalsePositiveFraction: unit1 must not be None')
        if unit2 is None:
            unit2 = self.getBestUnitMatch1(unit1)
            if unit2 is None or unit2 == -1:
                return 0

        if unit1 != -1 and unit2 != -1:
            a = self._matching_event_counts_12[unit1]
            if unit2 not in a:
                return 0
        else:
            return 0
        return 1 - self._compute_safe_frac(a[unit2], self._event_counts_1[unit1])

    def getFalseNegativeFraction(self, unit1, unit2=None):
        if unit1 is None:
            raise Exception('getFalsePositiveFraction: unit1 must not be None')
        if unit2 is None:
            unit2 = self.getBestUnitMatch1(unit1)
            if unit2 is None:
                return 0

        if unit1 != -1 and unit2 != -1:
            a = self._matching_event_counts_12[unit1]
            if unit2 not in a:
                return 0
        else:
            return 0
        return 1 - self._compute_safe_frac(a[unit2], self._event_counts_2[unit2])

    def computeCounts(self):
        if self._counts is None:
            self._do_counting(verbose=False)

    def _do_matching(self):
        self._event_counts_1 = dict()
        self._event_counts_2 = dict()
        self._matching_event_counts_12 = dict()
        self._best_match_units_12 = dict()
        self._matching_event_counts_21 = dict()
        self._best_match_units_21 = dict()
        self._unit_map12 = dict()
        self._unit_map21 = dict()

        sorting1 = self._sorting1
        sorting2 = self._sorting2
        unit1_ids = sorting1.get_unit_ids()
        unit2_ids = sorting2.get_unit_ids()
        N1 = len(unit1_ids)
        N2 = len(unit2_ids)

        # Compute events counts
        event_counts1 = np.zeros((N1)).astype(np.int64)
        for i1, u1 in enumerate(unit1_ids):
            times1 = sorting1.get_unit_spike_train(u1)
            event_counts1[i1] = len(times1)
            self._event_counts_1[u1] = len(times1)
        event_counts2 = np.zeros((N2)).astype(np.int64)
        for i2, u2 in enumerate(unit2_ids):
            times2 = sorting2.get_unit_spike_train(u2)
            event_counts2[i2] = len(times2)
            self._event_counts_2[u2] = len(times2)

        # Compute matching events
        matching_event_counts = np.zeros((N1, N2)).astype(np.int64)
        scores = np.zeros((N1, N2))
        for i1, u1 in enumerate(unit1_ids):
            times1 = sorting1.get_unit_spike_train(u1)
            for i2, u2 in enumerate(unit2_ids):
                times2 = sorting2.get_unit_spike_train(u2)
                num_matches = count_matching_events(times1, times2, delta=self._delta_tp)
                # test_unmatched=get_unmatched_times(times1,times2,delta=self._delta_tp)
                # print(u1,u2,'num_matches',num_matches,'num_unmatched',len(test_unmatched),'sum',num_matches+len(test_unmatched),'total',len(times1))
                matching_event_counts[i1, i2] = num_matches
                scores[i1, i2] = self._compute_agreement_score(num_matches, event_counts1[i1], event_counts2[i2])

        # Find best matches for spiketrains 1
        for i1, u1 in enumerate(unit1_ids):
            scores0 = scores[i1, :]
            self._matching_event_counts_12[u1] = dict()
            if len(scores0) > 0:
                if np.max(scores0) > 0:
                    inds0 = np.where(scores0 > 0)[0]
                    for i2 in inds0:
                        self._matching_event_counts_12[u1][unit2_ids[i2]] = matching_event_counts[i1, i2]
                    i2_best = np.argmax(scores0)
                    self._best_match_units_12[u1] = unit2_ids[i2_best]
                else:
                    self._best_match_units_12[u1] = -1
            else:
                self._best_match_units_12[u1] = -1

        # Find best matches for spiketrains 2
        for i2, u2 in enumerate(unit2_ids):
            scores0 = scores[:, i2]
            self._matching_event_counts_21[u2] = dict()
            if len(scores0) > 0:
                if np.max(scores0) > 0:
                    inds0 = np.where(scores0 > 0)[0]
                    for i1 in inds0:
                        self._matching_event_counts_21[u2][unit1_ids[i1]] = matching_event_counts[i1, i2]
                    i1_best = np.argmax(scores0)
                    self._best_match_units_21[u2] = unit1_ids[i1_best]
                else:
                    self._best_match_units_21[u2] = -1
            else:
                self._best_match_units_21[u2] = -1

        # Assign best matches
        [inds1, inds2] = linear_sum_assignment(-scores)
        inds1 = list(inds1)
        inds2 = list(inds2)
        # if len(unit2_ids) > 0:
        #     k2 = np.max(unit2_ids) + 1
        # else:
        #     k2 = 1
        for i1, u1 in enumerate(unit1_ids):
            if i1 in inds1:
                aa = inds1.index(i1)
                i2 = inds2[aa]
                u2 = unit2_ids[i2]
                if self.getAgreementFraction(u1, u2) > self._min_accuracy:
                    self._unit_map12[u1] = u2
                else:
                    self._unit_map12[u1] = -1
            else:
                # self._unit_map12[u1] = k2
                # k2 = k2+1
                self._unit_map12[u1] = -1
        # if len(unit1_ids) > 0:
        #     k1 = np.max(unit1_ids) + 1
        # else:
        #     k1 = 1
        for i2, u2 in enumerate(unit2_ids):
            if i2 in inds2:
                aa = inds2.index(i2)
                i1 = inds1[aa]
                u1 = unit1_ids[i1]
                if self.getAgreementFraction(u1, u2) > self._min_accuracy:
                    self._unit_map21[u2] = u1
                else:
                    self._unit_map21[u2] = -1
            else:
                # self._unit_map21[u2] = k1
                # k1 = k1+1
                self._unit_map21[u2] = -1

    def _do_counting(self, verbose=False):
        sorting1 = self._sorting1
        sorting2 = self._sorting2
        unit1_ids = sorting1.get_unit_ids()
        unit2_ids = sorting2.get_unit_ids()
        self._labels_st1 = dict()
        self._labels_st2 = dict()
        # N1 = len(unit1_ids)
        # N2 = len(unit2_ids)
        # Evaluate
        for u1 in unit1_ids:
            st1 = sorting1.get_unit_spike_train(u1)
            lab_st1 = np.array(['UNPAIRED'] * len(st1))
            self._labels_st1[u1] = lab_st1
        for u2 in unit2_ids:
            st2 = sorting2.get_unit_spike_train(u2)
            lab_st2 = np.array(['UNPAIRED'] * len(st2))
            self._labels_st2[u2] = lab_st2

        if verbose:
            print('Finding TP')
        for u1 in sorting1.get_unit_ids():
            if self.getMappedSorting1().getMappedUnitIds(u1) != -1:
                lab_st1 = self._labels_st1[u1]
                lab_st2 = self._labels_st2[self.getMappedSorting1().getMappedUnitIds(u1)]
                mapped_st = self.getMappedSorting1().get_unit_spike_train(u1)
                # from gtst: TP, TPO, TPSO, FN, FNO, FNSO
                for sp_i, n_sp in enumerate(sorting1.get_unit_spike_train(u1)):
                    id_sp = np.where((mapped_st > n_sp - self._delta_tp) & (mapped_st < n_sp + self._delta_tp))[0]
                    if len(id_sp) == 1:
                        lab_st1[sp_i] = 'TP'
                        lab_st2[id_sp] = 'TP'
            else:
                lab_st1 = np.array(['FN'] * len(sorting1.get_unit_spike_train(u1)))

        # find CL-CLO-CLSO
        if verbose:
            print('Finding CL')
        for u1 in sorting1.get_unit_ids():
            lab_st1 = self._labels_st1[u1]
            st1 = sorting1.get_unit_spike_train(u1)
            for l_gt, lab in enumerate(lab_st1):
                if lab == 'UNPAIRED':
                    for u2 in sorting2.get_unit_ids():
                        if u2 in self.getMappedSorting1().getMappedUnitIds() \
                                and self.getMappedSorting1().getMappedUnitIds(u1) != -1:
                            lab_st2 = self._labels_st2[u2]
                            st2 = sorting2.get_unit_spike_train(u2)

                            n_up = st1[l_gt]
                            id_sp = np.where((st2 > n_up - self._delta_tp) & (st2 < n_up + self._delta_tp))[0]
                            if len(id_sp) == 1 and lab_st2[id_sp] == 'UNPAIRED':
                                lab_st1[l_gt] = 'CL_' + str(u1) + '_' + str(u2)
                                lab_st2[id_sp] = 'CL_' + str(u2) + '_' + str(u1)
                                # if lab_st2[id_sp] == 'UNPAIRED':
                                #     lab_st2[id_sp] = 'CL_NP'

        if verbose:
            print('Finding FP and FN')
        for u1 in sorting1.get_unit_ids():
            lab_st1 = self._labels_st1[u1]
            for l_gt, lab in enumerate(lab_st1):
                if lab == 'UNPAIRED':
                    lab_st1[l_gt] = 'FN'

        for u2 in sorting2.get_unit_ids():
            lab_st2 = self._labels_st2[u2]
            for l_gt, lab in enumerate(lab_st2):
                if lab == 'UNPAIRED':
                    lab_st2[l_gt] = 'FP'

        TOT_ST1 = sum([len(sorting1.get_unit_spike_train(unit)) for unit in sorting1.get_unit_ids()])
        TOT_ST2 = sum([len(sorting2.get_unit_spike_train(unit)) for unit in sorting2.get_unit_ids()])
        total_spikes = TOT_ST1 + TOT_ST2
        TP = sum([len(np.where('TP' == self._labels_st1[unit])[0]) for unit in sorting1.get_unit_ids()])
        CL = sum(
            [len([i for i, v in enumerate(self._labels_st1[unit]) if 'CL' in v]) for unit in sorting1.get_unit_ids()])
        FN = sum([len(np.where('FN' == self._labels_st1[unit])[0]) for unit in sorting1.get_unit_ids()])
        FP = sum([len(np.where('FP' == self._labels_st2[unit])[0]) for unit in sorting2.get_unit_ids()])
        self.counts = {'TP': TP, 'CL': CL, 'FN': FN, 'FP': FP, 'TOT': total_spikes, 'TOT_ST1': TOT_ST1,
                       'TOT_ST2': TOT_ST2}

        if verbose:
            print('TP :', TP)
            print('CL :', CL)
            print('FN :', FN)
            print('FP :', FP)
            print('TOTAL: ', TOT_ST1, TOT_ST2, TP + CL + FN + FP)

    def _do_confusion(self):
        # def confusion_matrix(gtst, sst, pairs, plot_fig=True, xlabel=None, ylabel=None):
        '''

        Parameters
        ----------
        gtst
        sst
        pairs 1D array with paired sst to gtst

        Returns
        -------

        '''
        sorting1 = self._sorting1
        sorting2 = self._sorting2
        unit1_ids = sorting1.get_unit_ids()
        unit2_ids = sorting2.get_unit_ids()
        N1 = len(unit1_ids)
        N2 = len(unit2_ids)

        conf_matrix = np.zeros((N1 + 1, N2 + 1), dtype=int)
        idxs_matched = np.where(np.array(self.getMappedSorting1().getMappedUnitIds()) != -1)
        if len(idxs_matched) > 0:
            idxs_matched = idxs_matched[0]
        idxs_unmatched = np.where(np.array(self.getMappedSorting1().getMappedUnitIds()) == -1)
        if len(idxs_unmatched) > 0:
            idxs_unmatched = idxs_unmatched[0]
        unit_map_matched = np.array(self.getMappedSorting1().getMappedUnitIds())[idxs_matched]

        st1_idxs = np.append(np.array(sorting1.get_unit_ids())[idxs_matched],
                             np.array(sorting1.get_unit_ids())[idxs_unmatched])
        st2_matched = unit_map_matched
        st2_unmatched = []

        for u_i, u1 in enumerate(np.array(sorting1.get_unit_ids())[idxs_matched]):
            lab_st1 = self._labels_st1[u1]
            tp = len(np.where('TP' == lab_st1)[0])
            conf_matrix[u_i, u_i] = int(tp)
            for u2 in sorting2.get_unit_ids():
                lab_st2 = self._labels_st2[u2]
                cl_str = str(u1) + '_' + str(u2)
                cl = len([i for i, v in enumerate(lab_st1) if 'CL' in v and cl_str in v])
                if cl != 0:
                    st_p = np.where(u2 == unit_map_matched)
                    conf_matrix[u_i, st_p] = int(cl)
            fn = len(np.where('FN' == lab_st1)[0])
            conf_matrix[u_i, -1] = int(fn)

        for u_i, u1 in enumerate(np.array(sorting1.get_unit_ids())[idxs_unmatched]):
            lab_st1 = self._labels_st1[u1]
            fn = len(np.where('FN' == lab_st1)[0])
            conf_matrix[u_i + len(idxs_matched), -1] = int(fn)

        for _, u2 in enumerate(sorting2.get_unit_ids()):
            lab_st2 = self._labels_st2[u2]
            fp = len(np.where('FP' == lab_st2)[0])
            st_p = np.where(u2 == unit_map_matched)[0]
            if len(st_p) != 0:
                conf_matrix[-1, st_p] = int(fp)
            else:
                st2_unmatched.append(int(u2))
                conf_matrix[-1, len(idxs_matched) + len(st2_unmatched) - 1] = int(fp)

        self._confusion_matrix = conf_matrix
        st2_idxs = np.append(st2_matched, st2_unmatched)

        return st1_idxs, st2_idxs

    @staticmethod
    def compareSpikeTrains(spiketrain1, spiketrain2, delta_tp=10, verbose=False):
        lab_st1 = np.array(['UNPAIRED'] * len(spiketrain1))
        lab_st2 = np.array(['UNPAIRED'] * len(spiketrain2))

        if verbose:
            print('Finding TP')
        # from gtst: TP, TPO, TPSO, FN, FNO, FNSO
        for sp_i, n_sp in enumerate(spiketrain1):
            id_sp = np.where((spiketrain2 > n_sp - delta_tp) & (spiketrain2 < n_sp + delta_tp))[0]
            if len(id_sp) == 1:
                lab_st1[sp_i] = 'TP'
                lab_st2[id_sp] = 'TP'

        if verbose:
            print('Finding FP and FN')
        for l_gt, lab in enumerate(lab_st1):
            if lab == 'UNPAIRED':
                lab_st1[l_gt] = 'FN'

        for l_gt, lab in enumerate(lab_st2):
            if lab == 'UNPAIRED':
                lab_st2[l_gt] = 'FP'

        return lab_st1, lab_st2


class MappedSortingExtractor(se.SortingExtractor):
    def __init__(self, sorting, unit_map):
        se.SortingExtractor.__init__(self)
        self._sorting = sorting
        self._unit_map = unit_map
        self._unit_ids = list(self._unit_map.keys())

    def get_unit_ids(self, unit_ids=None):
        if unit_ids is None:
            return self._unit_ids
        else:
            return self._unit_ids[unit_ids]

    def getMappedUnitIds(self, unit_ids=None):
        if unit_ids is None:
            return list(self._unit_map.values())
        elif isinstance(unit_ids, (int, np.integer)):
            return self._unit_map[unit_ids]
        else:
            return list([self._unit_map[u] for u in self._unit_ids if u in unit_ids])

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        unit2 = self._unit_map[unit_id]
        if unit2 != -1:
            return self._sorting.get_unit_spike_train(unit2, start_frame=start_frame, end_frame=end_frame)
        else:
            print(unit_id, " is not matched!")
            return None

# for troubleshooting


def get_unmatched_times(times1, times2, *, delta):
    times1 = np.array(times1)
    times2 = np.array(times2)
    times_concat = np.concatenate((times1, times2))
    membership = np.concatenate((np.ones(times1.shape) * 1, np.ones(times2.shape) * 2))
    indices = times_concat.argsort()
    times_concat_sorted = times_concat[indices]
    membership_sorted = membership[indices]
    diffs = times_concat_sorted[1:] - times_concat_sorted[:-1]
    unmatched_inds = 1 + np.where((diffs[1:] > delta) & (diffs[:-1] > delta) & (membership_sorted[1:-1] == 1))[0]
    if (diffs[0] > delta) and (membership_sorted[0] == 1):
        unmatched_inds = np.concatenate(([0], unmatched_inds))
    if (diffs[-1] > delta) and (membership_sorted[-1] == 1):
        unmatched_inds = np.concatenate((unmatched_inds, [len(membership_sorted) - 1]))
    return times_concat_sorted[unmatched_inds]


def count_matching_events(times1, times2, delta=10):
    times_concat = np.concatenate((times1, times2))
    membership = np.concatenate((np.ones(times1.shape) * 1, np.ones(times2.shape) * 2))
    indices = times_concat.argsort()
    times_concat_sorted = times_concat[indices]
    membership_sorted = membership[indices]
    diffs = times_concat_sorted[1:] - times_concat_sorted[:-1]
    inds = np.where((diffs <= delta) & (membership_sorted[0:-1] != membership_sorted[1:]))[0]
    if (len(inds) == 0):
        return 0
    inds2 = np.where(inds[:-1] + 1 != inds[1:])[0]
    return len(inds2) + 1


def confusion_matrix(gtst, sst, pairs, plot_fig=True, xlabel=None, ylabel=None):
    '''

    Parameters
    ----------
    gtst
    sst
    pairs 1D array with paired sst to gtst

    Returns
    -------

    '''
    conf_matrix = np.zeros((len(gtst) + 1, len(sst) + 1), dtype=int)
    idxs_pairs_clean = np.where(pairs != -1)
    idxs_pairs_dirty = np.where(pairs == -1)
    pairs_clean = pairs[idxs_pairs_clean]
    gtst_clean = np.array(gtst)[idxs_pairs_clean]
    gtst_extra = np.array(gtst)[idxs_pairs_dirty]

    # gtst_idxs = np.append(idxs_pairs_clean, idxs_pairs_dirty)
    # sst_idxs = pairs_clean
    sst_extra = []

    for gt_i, gt in enumerate(gtst_clean):
        if gt.annotations['paired']:
            tp = len(np.where('TP' == gt.annotations['labels'])[0])
            conf_matrix[gt_i, gt_i] = int(tp)
            for st_i, st in enumerate(sst):
                cl_str = str(gt_i) + '_' + str(st_i)
                cl = len([i for i, v in enumerate(gt.annotations['labels']) if 'CL' in v and cl_str in v])
                if cl != 0:
                    st_p = np.where(st_i == pairs_clean)
                    conf_matrix[gt_i, st_p] = int(cl)
        fn = len(np.where('FN' == gt.annotations['labels'])[0])
        conf_matrix[gt_i, -1] = int(fn)
    for gt_i, gt in enumerate(gtst_extra):
        fn = len(np.where('FN' == gt.annotations['labels'])[0])
        conf_matrix[gt_i + len(gtst_clean), -1] = int(fn)
    for st_i, st in enumerate(sst):
        fp = len(np.where('FP' == st.annotations['labels'])[0])
        st_p = np.where(st_i == pairs_clean)[0]
        if len(st_p) != 0:
            conf_matrix[-1, st_p] = fp
        else:
            sst_extra.append(int(st_i))
            conf_matrix[-1, len(pairs_clean) + len(sst_extra) - 1] = fp

    ax = None
    return conf_matrix, ax


def compute_performance(SC):
    counts = SC.counts

    tp_rate = float(counts['TP']) / counts['TOT_ST1'] * 100
    cl_rate = float(counts['CL']) / counts['TOT_ST1'] * 100
    fn_rate = float(counts['FN']) / counts['TOT_ST1'] * 100
    fp_st1 = float(counts['FP']) / counts['TOT_ST1'] * 100
    fp_st2 = float(counts['FP']) / counts['TOT_ST2'] * 100

    accuracy = tp_rate / (tp_rate + fn_rate + fp_st1) * 100
    sensitivity = tp_rate / (tp_rate + fn_rate) * 100
    miss_rate = fn_rate / (tp_rate + fn_rate) * 100
    precision = tp_rate / (tp_rate + fp_st1) * 100
    false_discovery_rate = fp_st1 / (tp_rate + fp_st1) * 100

    print('PERFORMANCE: \n')
    print('TP: ', tp_rate, ' %')
    print('CL: ', cl_rate, ' %')
    print('FN: ', fn_rate, ' %')
    print('FP (%ST1): ', fp_st1, ' %')
    print('FP (%ST2): ', fp_st2, ' %')

    print('\nACCURACY: ', accuracy, ' %')
    print('SENSITIVITY: ', sensitivity, ' %')
    print('MISS RATE: ', miss_rate, ' %')
    print('PRECISION: ', precision, ' %')
    print('FALSE DISCOVERY RATE: ', false_discovery_rate, ' %')

    performance = {'tp': tp_rate, 'cl': cl_rate, 'fn': fn_rate, 'fp_st1': fp_st1, 'fp_st2': fp_st2,
                   'accuracy': accuracy, 'sensitivity': sensitivity, 'precision': precision, 'miss_rate': miss_rate,
                   'false_disc_rate': false_discovery_rate}

    return performance



import os
from mountaintools import client as mt
import mlprocessors as mlpr
import shutil


def install_kilosort2(repo, commit):
    spikeforest_alg_install_path = os.getenv('SPIKEFOREST_ALG_INSTALL_PATH', os.getenv('HOME') + '/spikeforest_algs')
    key = dict(
        alg='kilosort2',
        repo=repo,
        commit=commit
    )
    source_path = spikeforest_alg_install_path + '/kilosort2_' + commit
    if os.path.exists(source_path):
        # The dir hash method does not seem to be working for some reason here
        # hash0 = mt.computeDirHash(source_path)
        # if hash0 == mt.getValue(key=key):
        #     print('Kilosort2 is already auto-installed.')
        #     return source_path

        a = mt.loadObject(path=source_path + '/spikeforest.json')
        if a:
            if mt.sha1OfObject(a) == mt.sha1OfObject(key):
                print('Kilosort2 is already auto-installed.')
                return source_path

        print('Removing directory: {}'.format(source_path))
        shutil.rmtree(source_path)

    script = """
    #!/bin/bash
    set -e

    git clone {repo} {source_path}
    cd {source_path}
    git checkout {commit}
    """.format(repo=repo, commit=commit, source_path=source_path)
    ss = mlpr.ShellScript(script=script)
    ss.start()
    retcode = ss.wait()
    if retcode != 0:
        raise Exception('Install script returned a non-zero exit code/')

    compile_gpu = mlpr.ShellScript(script="""
    function compile_gpu

    try
        run('mexGPUall.m');
    catch
        disp('Problem running mexGPUall.');
        disp(lasterr());
        exit(-1)
    end;
    exit(0)
    """)
    compile_gpu.write(script_path=source_path + '/CUDA/compile_gpu.m')

    script = """
    #!/bin/bash
    set -e

    cd {source_path}/CUDA
    matlab -nodisplay -nosplash -r "compile_gpu"
    """.format(source_path=source_path)
    ss = mlpr.ShellScript(script=script)
    ss.start()
    retcode = ss.wait()
    if retcode != 0:
        raise Exception('Compute gpu script returned a non-zero exit code.')

    # The dir hash method does not seem to be working for some reason here
    # hash0 = mt.computeDirHash(source_path)
    # mt.setValue(key=key, value=hash0)
    mt.saveObject(object=key, dest_path=source_path + '/spikeforest.json')

    return source_path



from .mountainsort4 import MountainSort4, MountainSort4TestError
from .spyking_circus import SpykingCircus
from .yass import YASS
from .yass1 import YASS1
from .ironclust import IronClust
from .kilosort import KiloSort
from .kilosort2 import KiloSort2
from .herdingspikes2 import HerdingSpikes2
from .jrclust import JRClust
from .klusta import Klusta
from .tridesclous import Tridesclous
from .waveclus import Waveclus



from spikeextractors import SortingExtractor
import numpy as np
from pathlib import Path
import h5py


class SpykingCircusSortingExtractor(SortingExtractor):
    def __init__(self, spykingcircus_folder):
        SortingExtractor.__init__(self)
        spykingcircus_folder = Path(spykingcircus_folder)
        files = spykingcircus_folder.iterdir()
        results = None
        for f in files:
            if 'result.hdf5' in str(f):
                results = f
            if 'result-merged.hdf5' in str(f):
                results = f
                break
        if results is None:
            raise Exception(spykingcircus_folder, " is not a spyking circus folder. No file found: result.hdf5 or result-merged.hdf5.")
        f_results = h5py.File(results)
        self._spiketrains = []
        self._unit_ids = []
        for temp in f_results['spiketimes'].keys():
            self._spiketrains.append(f_results['spiketimes'][temp].value)
            self._unit_ids.append(int(temp.split('_')[-1]))

    def get_unit_ids(self):
        return list(self._unit_ids)

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        times = self._spiketrains[self.get_unit_ids().index(unit_id)]
        inds = np.where((start_frame <= times) & (times < end_frame))
        return times[inds]

    @staticmethod
    def write_sorting(sorting, save_path):
        save_path = Path(save_path)
        if save_path.is_dir():
            save_path = save_path / 'data.result.hdf5'
        elif save_path.suffix == '.hdf5':
            if not str(save_path).endswith('result.hdf5') or not str(save_path).endswith('result-merged.hdf5'):
                raise AttributeError("'save_path' is either a folder or an hdf5 file "
                                     "ending with 'result.hdf5' or 'result-merged.hdf5")
        else:
            save_path.mkdir()
            save_path = save_path / 'data.result.hdf5'
        F = h5py.File(save_path, 'w')
        spiketimes = F.create_group('spiketimes')

        for id in sorting.get_unit_ids():
            spiketimes.create_dataset('tmp_' + str(id), data=sorting.get_unit_spike_train(id))



from matplotlib import pyplot as plt
import numpy as np
import spikeforest


class CrossCorrelogramsWidget:
    def __init__(self, max_samples=None, auto=True, *, sorting, samplerate, unit_ids=None, _figure=None, _axs=None):
        self._SX = sorting
        self._unit_ids = unit_ids
        self._figure = _figure
        self._axs = _axs
        if self._figure is not None:
            self._axs = self._figure.axes
        elif self._axs is not None:
            self._axs = self._axs
        self._samplerate = samplerate
        self.auto = auto
        self.max_samples = max_samples
        self.max_dt_msec = 50
        self.bin_size_msec = 2
        self.max_dt_tp = self.max_dt_msec * self._samplerate / 1000
        self.bin_size_tp = self.bin_size_msec * self._samplerate / 1000

    def plot(self):
        if self.auto:
            self._do_plot()
        else:
            self._do_plot_matrix()

    def figure(self):
        return self._figure

    def _do_plot_matrix(self):
        units = self._unit_ids
        if units is None:
            units = self._SX.get_unit_ids()
        nrows = ncols = len(units)
        f, axs = plt.subplots(nrows, ncols, figsize=(3 * ncols + 0.1, 3 * nrows + 0.1))
        self._figure = f
        for i1, unit1 in enumerate(units):
            times1 = self._SX.get_unit_spike_train(unit_id=unit1)
            for i2, unit2 in enumerate(units):
                times2 = self._SX.get_unit_spike_train(unit_id=unit2)
                if i1 == i2:
                    (bin_counts, bin_edges) = compute_crosscorrelogram(times1, max_dt_tp=self.max_dt_tp, bin_size_tp=self.bin_size_tp, max_samples=self.max_samples)
                else:
                    (bin_counts, bin_edges) = compute_crosscorrelogram(times1, times2, max_dt_tp=self.max_dt_tp, bin_size_tp=self.bin_size_tp, max_samples=self.max_samples)
                item = dict(
                    title="{} -> {}".format(unit1, unit2),
                    bin_counts=bin_counts,
                    bin_edges=bin_edges
                )
                self._plot_correlogram(axs[i1, i2], **item)

    def _do_plot(self):
        units = self._unit_ids
        if units is None:
            units = self._SX.get_unit_ids()
        list = []
        for unit in units:
            times = self._SX.get_unit_spike_train(unit_id=unit)
            (bin_counts, bin_edges) = compute_autocorrelogram(times, max_dt_tp=self.max_dt_tp, bin_size_tp=self.bin_size_tp)
            item = dict(
                title=str(unit),
                bin_counts=bin_counts,
                bin_edges=bin_edges
            )
            list.append(item)
        with plt.rc_context({'axes.edgecolor': 'gray'}):
            self._plot_correlograms_multi(list)

    def _plot_correlogram(self, ax, *, bin_counts, bin_edges, title=''):
        wid = (bin_edges[1] - bin_edges[0]) * 1000
        ax.bar(x=bin_edges[0:-1] * 1000, height=bin_counts, width=wid, color='gray', align='edge')
        ax.set_xlabel('dt (msec)')
        ax.set_xticks([])
        ax.set_yticks([])
        if title:
            ax.set_title(title, color='gray')

    def _plot_correlograms_multi(self, list, *, ncols=5, **kwargs):
        nrows = int(np.ceil(len(list) / ncols))
        if (self._figure is None) & (self._axs is None):
            f, axs = plt.subplots(nrows, ncols, figsize=(3 * ncols + 0.1, 3 * nrows + 0.1))
            self._figure = f
        for i, item in enumerate(list):
            ax = plt.subplot(nrows, ncols, i + 1)
            self._plot_correlogram(ax, **item, **kwargs)
        self._axs = axs


def compute_crosscorrelogram(x, y=None, *, max_dt_tp, bin_size_tp, max_samples=None):
    if y is None:
        y = x
        auto = True
    else:
        auto = False
    if max_samples is not None:
        if max_samples < len(x):
            x = np.random.choice(x, size=max_samples, replace=False)
        if max_samples < len(y):
            y = np.random.choice(y, size=max_samples, replace=False)

    bin_start = -max_dt_tp
    bin_stop = max_dt_tp
    bin_edges = np.arange(start=bin_start, stop=bin_stop + bin_size_tp, step=bin_size_tp)
    counts = np.zeros(len(bin_edges) - 1)
    nbins = len(counts)
    x = np.sort(x)
    lx = len(x) - 1
    y = np.sort(y)
    for yspk in y:
        xlo = np.searchsorted(x, yspk + bin_start)  # find first spike of x within window relative to yspk
        xhi = np.searchsorted(x[xlo:], yspk + bin_stop, side='right') + xlo  # find the last one
        xlo = max(0, xlo)  # make sure they're in bounds
        xhi = min(lx, xhi)
        for i in np.arange(xlo, xhi):
            yspkdiff = x[i] - yspk  # get time difference between events
            bin = min(np.searchsorted(bin_edges, yspkdiff), nbins)  # get bin of time difference
            if ((yspkdiff == 0) & auto) | (bin == 0):  # check bounds and exclude same event for autocorrelation
                continue
            counts[bin - 1] += 1  # add to counts
    return (counts, bin_edges)


def compute_autocorrelogram(times, *, max_dt_tp, bin_size_tp, max_samples=None):
    num_bins_left = int(max_dt_tp / bin_size_tp)  # number of bins to the left of the origin
    L = len(times)  # number of events
    times2 = np.sort(times)  # the sorted times
    step = 1  # This is the index step between an event and the next one to compare
    candidate_inds = np.arange(L)  # These are the events we are going to consider
    if max_samples is not None:
        if len(candidate_inds) > max_samples:
            candidate_inds = np.random.choice(candidate_inds, size=max_samples, replace=False)
    vals_list = []  # A list of all offsets we have accumulated
    while True:
        candidate_inds = candidate_inds[
            candidate_inds + step < L]  # we only consider events that are within workable range
        candidate_inds = candidate_inds[times2[candidate_inds + step] - times2[
            candidate_inds] <= max_dt_tp]  # we only consider event-pairs that are within max_dt_tp apart
        if len(candidate_inds) > 0:  # if we have some events to consider
            vals = times2[candidate_inds + step] - times2[candidate_inds]
            vals_list.append(vals)  # add to the autocorrelogram
            vals_list.append(-vals)  # keep it symmetric
        else:
            break  # no more to consider
        step += 1
    if len(vals_list) > 0:  # concatenate all the values
        all_vals = np.concatenate(vals_list)
    else:
        all_vals = np.array([])
    aa = np.arange(-num_bins_left, num_bins_left + 1) * bin_size_tp
    all_vals = np.sign(all_vals) * (np.abs(
        all_vals) - bin_size_tp * 0.00001)  # a trick to make the histogram symmetric due to differences in rounding for positive and negative, i suppose
    bin_counts, bin_edges = np.histogram(all_vals, bins=aa)
    return (bin_counts, bin_edges)


def test_crosscorrelogram():
    # test 1
    offset = 2
    ans = spikeforest.spikewidgets.widgets.crosscorrelogramswidget.compute_crosscorrelogram(np.arange(1, 400, 25),
                                                                                            np.arange(1 + offset, 400, 25),
                                                                                            max_dt_tp=20,
                                                                                            bin_size_tp=0.1)
    ind = np.where(ans[0])
    assert ((ans[1][ind[0]] < offset ) & (ans[1][ind[0] + 1] <= offset ))[0]

    # test 2
    case = np.arange(1, 400, 15)
    counts1, edges1 = spikeforest.spikewidgets.widgets.crosscorrelogramswidget.compute_autocorrelogram(times, max_dt_tp=20, bin_size_tp=0.1)
    counts2, edges2 = spikeforest.spikewidgets.widgets.crosscorrelogramswidget.compute_crosscorrelogram(times, max_dt_tp=20, bin_size_tp=0.1)

    roughly_equals(counts1, counts2)


def roughly_equals(a, b):
    # TODO: implement
    return True




# %% Change working directory from the workspace root to the ipynb file location. Turn this addition off with the DataScience.changeDirOnImportExport setting
import os
try:
    os.chdir(os.path.join(os.getcwd(), 'ironclust_test'))
    print(os.getcwd())
except:
    pass

# %% python magic
# get_ipython().run_line_magic('load_ext', 'autoreload')
# get_ipython().run_line_magic('autoreload', '2')

# %%
import spikeforest_analysis as sa
from spikeforest_analysis.compare_sortings_with_truth import GenSortingComparisonTable
# from mountaintools import client as ca
import spikeextractors as se
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor, example_datasets
import os
from spikeforest_analysis.sorters import ironclust
from spikeforest import SortingComparison
from spikeforest import spikewidgets as sw
from spikesorters import IronClust

# %%SortingComparisonyass_test1/recording/raw.mda
""" recording_path = 'irc_test1'
if not os.path.isdir(recording_path):
    os.mkdir(recording_path)
rx, sx = example_datasets.yass_example(set_id=1) """

# %%
recording_path = 'kbucket://15734439d8cf/groundtruth/visapy_mea/set1'
firings_true = recording_path + '/firings_true.mda'
# recording_path = recording_path+'/recording'
# SFMdaRecordingExtractor.write_recording(
# recording=rx, save_path=recording_path)
# SFMdaSortingExtractor.write_sorting(
# sorting=sx, save_path=firings_true)
result_path = os.path.abspath('test_irc')
if not os.path.exists(result_path):
    os.mkdir(result_path)
firings_out = os.path.join(result_path, 'firings_out.mda')
print('output stored in: ', firings_out)

IronClust.execute(
    recording_dir=recording_path,
    firings_out=firings_out,
    prm_template_name='static',
    detect_sign=-1,
    adjacency_radius=50,
    _container=None,
    _force_run=True,
    _keep_temp_files=True
)
#firings_out = recording_path+'/firings_out.mda'
assert os.path.exists(firings_out)


# %%
print('recording: {}'.format(recording_path))
print('firings_out: {}'.format(firings_out))
print('firings_true: {}'.format(firings_true))

# %%
GenSortingComparisonTable.execute(
    firings=firings_out,
    firings_true=firings_true,
    units_true=[],
    json_out=os.path.join(recording_path, 'out.json'),
    html_out=os.path.join(recording_path, 'out.html'),
    _container=None
)


# %%


# %%



import vdomr as vd
import time
import multiprocessing
import sys
from .stdoutsender import StdoutSender
import mtlogging
import numpy as np
from copy import deepcopy
import spikeforest_analysis as sa
from spikeforest import EfficientAccessRecordingExtractor
import spikeforestwidgets as SFW
import json
from matplotlib import pyplot as plt
import mlprocessors as mlpr
import uuid


class TemplatesView(vd.Component):
    def __init__(self, *, context, opts=None, prepare_result):
        vd.Component.__init__(self)
        self._sorting_context = context
        self._recording_context = context.recordingContext()
        self._size = (100, 100)
        units = prepare_result['units']
        self._templates_widget = TemplatesWidget(units=units)
        self._update_selected()
        self._templates_widget.onSelectedUnitIdsChanged(lambda: self._sorting_context.setSelectedUnitIds(self._templates_widget.selectedUnitIds()))
        self._sorting_context.onSelectedUnitIdsChanged(lambda: self._templates_widget.setSelectedUnitIds(self._sorting_context.selectedUnitIds()))
        self._templates_widget.onCurrentUnitIdChanged(lambda: self._sorting_context.setCurrentUnitId(self._templates_widget.currentUnitId()))
        self._sorting_context.onCurrentUnitIdChanged(lambda: self._templates_widget.setCurrentUnitId(self._sorting_context.currentUnitId()))
        # self._templates_widget.setSize(self._size)
        self.refresh()

    @staticmethod
    def prepareView(context, opts):
        sorting_context = context
        recording_context = context.recordingContext()
        recording_context.initialize()
        sorting_context.initialize()
        earx = EfficientAccessRecordingExtractor(recording=recording_context.recordingExtractor())
        sorting = sorting_context.sortingExtractor()
        unit_ids = sorting.get_unit_ids()
        print('***** Computing unit templates...')
        # templates = compute_unit_templates(recording=earx, sorting=sorting, unit_ids=unit_ids)
        templates = ComputeUnitTemplates.execute(recording=earx, sorting=sorting, unit_ids=unit_ids, templates_out=True).outputs['templates_out']
        print('*****')
        return dict(units=[
            dict(
                template=templates[:, :, i],
                unit_id=unit_id
            )
            for i, unit_id in enumerate(unit_ids)
        ])

    def setSize(self, size):
        if self._size != size:
            self._size = size
        # if self._templates_widget:
        #    self._templates_widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Templates'

    def render(self):
        return self._templates_widget

    def _update_selected(self):
        self._templates_widget.setSelectedUnitIds(self._sorting_context.selectedUnitIds())
        self._templates_widget.setCurrentUnitId(self._sorting_context.currentUnitId())


def get_random_spike_waveforms(*, recording, sorting, unit, snippet_len, max_num, channels=None):
    st = sorting.get_unit_spike_train(unit_id=unit)
    num_events = len(st)
    if num_events > max_num:
        event_indices = np.random.choice(range(num_events), size=max_num, replace=False)
    else:
        event_indices = range(num_events)

    spikes = recording.get_snippets(reference_frames=st[event_indices].astype(int), snippet_len=snippet_len, channel_ids=channels)
    if len(spikes) > 0:
        spikes = np.dstack(tuple(spikes))
    else:
        spikes = np.zeros((recording.get_num_channels(), snippet_len, 0))
    return spikes


def compute_unit_templates(*, recording, sorting, unit_ids, snippet_len=50, max_num=100, channels=None):
    M = len(recording.get_channel_ids())
    T = snippet_len
    K = len(unit_ids)
    ret = np.zeros((M, T, K))
    for i, unit in enumerate(unit_ids):
        print('Unit {} of {} (id={})'.format(i, len(unit_ids), unit))
        waveforms = get_random_spike_waveforms(recording=recording, sorting=sorting, unit=unit, snippet_len=snippet_len, max_num=max_num, channels=None)
        template = np.median(waveforms, axis=2)
        ret[:, :, i] = template

    return ret


class ComputeUnitTemplates(mlpr.Processor):
    NAME = 'ComputeUnitTemplates'
    VERSION = '0.1.4'
    recording = mlpr.Input()
    sorting = mlpr.Input()
    templates_out = mlpr.OutputArray()

    def run(self):
        templates = compute_unit_templates(recording=self.recording, sorting=self.sorting, unit_ids=self.sorting.get_unit_ids())  # pylint: disable=no-member
        print('Saving templates...', self.templates_out)
        np.save(self.templates_out, templates)


class TemplatesWidget(vd.Component):
    def __init__(self, *, units):
        vd.Component.__init__(self)
        y_scale_factor = _compute_initial_y_scale_factor(units)
        self._current_unit_id = None
        self._selected_unit_ids = []
        self._selected_unit_ids_changed_handlers = []
        self._current_unit_id_changed_handlers = []
        self._widgets = [
            TemplateWidget(
                template=unit['template'],
                unit_id=unit['unit_id'],
                y_scale_factor=y_scale_factor,
                size=(150, 300)
            )
            for unit in units
        ]
        self._uuid = str(uuid.uuid4())
        vd.register_callback(self._uuid + '_box_clicked', self._handle_box_clicked)
        vd.devel.loadBootstrap()

    def setSelectedUnitIds(self, ids):
        if ids is None:
            ids = []
        ids = sorted([int(id) for id in ids])
        if ids == self._selected_unit_ids:
            return
        self._selected_unit_ids = ids
        ids_set = set(ids)
        for W in self._widgets:
            W.setSelected(W.unitId() in ids_set)
        for handler in self._selected_unit_ids_changed_handlers:
            handler()

    def selectedUnitIds(self):
        return deepcopy(self._selected_unit_ids)

    def onSelectedUnitIdsChanged(self, handler):
        self._selected_unit_ids_changed_handlers.append(handler)

    def setCurrentUnitId(self, id):
        if id == self._current_unit_id:
            return
        self._current_unit_id = id
        for W in self._widgets:
            W.setCurrent(W.unitId() == id)
        for handler in self._current_unit_id_changed_handlers:
            handler()

    def currentUnitId(self):
        return self._current_unit_id

    def onCurrentUnitIdChanged(self, handler):
        self._current_unit_id_changed_handlers.append(handler)

    def _handle_box_clicked(self, *, index, ctrlKey):
        W = self._widgets[int(index)]
        if ctrlKey:
            a = self.selectedUnitIds()
            if W.unitId() in a:
                a.remove(W.unitId())
            else:
                a.append(W.unitId())
            self.setSelectedUnitIds(a)
        else:
            self.setSelectedUnitIds([W.unitId()])
            self.setCurrentUnitId(W.unitId())

    def render(self):
        box_style = dict(float='left')
        boxes = [
            vd.div(W, id=self._uuid + '-{}'.format(i), style=box_style)
            for i, W in enumerate(self._widgets)
        ]
        div = vd.div(
            boxes,
            style=dict(overflow='auto', height='100%')
        )
        return div

    def postRenderScript(self):
        js = """
      for (let i=0; i<{num_widgets}; i++) {
          (function(index) {
            let elmt = document.getElementById('{uuid}-'+index);
            if (elmt) {
                elmt.onclick=function(evt) {
                    window.vdomr_invokeFunction('{box_clicked_callback_id}', [], {index:index, ctrlKey:evt.ctrlKey});
                }
            }
          })(i);
      }
      """
        js = js.replace('{num_widgets}', str(len(self._widgets)))
        js = js.replace('{uuid}', self._uuid)
        js = js.replace('{box_clicked_callback_id}', self._uuid + '_box_clicked')
        return js


def _compute_initial_y_scale_factor(units):
    templates_list = [unit['template'] for unit in units]
    all_templates = np.stack(templates_list)
    all_abs_vals = np.abs(all_templates.ravel())
    vv = np.percentile(all_abs_vals, 99)
    if vv > 0:
        return 1 / (2 * vv)
    else:
        return 1


class TemplateWidget(vd.Component):
    def __init__(self, *, unit_id, template, y_scale_factor=None, size=(200, 200)):
        vd.Component.__init__(self)
        self._unit_id = unit_id
        self._selected = False
        self._current = False
        self._size = size
        self._size_plot = (size[0] - 20, size[1] - 40)
        self._plot = SFW.TemplateWidget(template=template, size=self._size_plot)
        self._plot.setYScaleFactor(y_scale_factor)

    def setSelected(self, val):
        if self._selected == val:
            return
        self._selected = val
        self.refresh()

    def setCurrent(self, val):
        if self._current == val:
            return
        self._current = val
        self.refresh()

    def unitId(self):
        return self._unit_id

    def render(self):
        style_plot = {'position': 'relative', 'border': 'solid 1px black', 'background-color': 'white', 'left': '10px', 'bottom': '10px', 'width': '{}px'.format(self._size_plot[0]), 'height': '{}px'.format(self._size_plot[1])}
        style1 = {'width': '{}px'.format(self._size[0]), 'height': '{}px'.format(self._size[1])}
        if self._selected:
            style1['background-color'] = 'rgb(240,240,200)'
            style1['border'] = 'solid 1px gray'
            if self._current:
                style1['background-color'] = 'rgb(220,220,180)'
        if self._current:
            style1['border'] = 'solid 2px rgb(150,50,50)'
        return vd.div(
            vd.p('Unit {}'.format(self._unit_id), style={'text-align': 'center'}),
            vd.div(self._plot, style=style_plot, size=()),
            style=style1
        )



import mlprocessors as mlpr
import os
import time
import numpy as np
from os.path import join
from subprocess import Popen, PIPE
import shlex
import random
import string
import shutil
from spikeforest import mdaio
import spikeextractors as se
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
import sys
import shlex
import traceback
from .install_kilosort2 import install_kilosort2
# from ..ironclust.install_ironclust import install_ironclust
# import h5py


class KiloSort2(mlpr.Processor):
    """
    KiloSort2 wrapper for SpikeForest framework
      written by J. James Jun, May 7, 2019
      modified from `spiketoolkit/sorters/Kilosort`
      to be made compatible with SpikeForest

    [Prerequisite]
    1. MATLAB (Tested on R2018b)
    2. CUDA Toolkit v9.1

    [Installation instruction in SpikeForest environment]
    1. Run `git clone https://github.com/alexmorley/Kilosort2.git`
      Kilosort2 currently doesn't work on tetrodes and low-channel count probes (as of May 7, 2019).
      Clone from Alex Morley's repository that fixed these issues.
      Original Kilosort2 code can be obtained from `https://github.com/MouseLand/Kilosort2.git`
    2. (optional) If Alex Morley's latest version doesn't work with SpikeForest, run
        `git checkout 43cbbfff89b9c88cdeb147ffd4ac35bfde9c7956`
    3. In Matlab, run `CUDA/mexGPUall` to compile all CUDA codes
    4. Add `KILOSORT2_PATH=...` in your .bashrc file.
    """

    NAME = 'KiloSort2'
    VERSION = '0.3.3'  # wrapper VERSION
    ADDITIONAL_FILES = ['*.m']
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS']
    CONTAINER = None

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    channels = mlpr.IntegerListParameter(
        description='List of channels to use.', optional=True, default=[])
    firings_out = mlpr.Output('Output firings file')

    detect_sign = mlpr.IntegerParameter(default=-1, optional=True,
                                        description='Use -1 or 1, depending on the sign of the spikes in the recording')
    adjacency_radius = mlpr.FloatParameter(default=30, optional=True,
                                           description='Use -1 to include all channels in every neighborhood')
    detect_threshold = mlpr.FloatParameter(
        optional=True, default=6, description='')
    # prm_template_name=mlpr.StringParameter(optional=False,description='TODO')
    freq_min = mlpr.FloatParameter(
        optional=True, default=150, description='Use 0 for no bandpass filtering')
    freq_max = mlpr.FloatParameter(
        optional=True, default=6000, description='Use 0 for no bandpass filtering')
    merge_thresh = mlpr.FloatParameter(
        optional=True, default=0.98, description='TODO')
    pc_per_chan = mlpr.IntegerParameter(
        optional=True, default=3, description='TODO')
    minFR = mlpr.FloatParameter(default=1 / 50, optional=True,
                                description='minimum spike rate (Hz), if a cluster falls below this for too long it gets removed')

    @staticmethod
    def install():
        print('Auto-installing kilosort.')
        return install_kilosort2(
            repo='https://github.com/alexmorley/Kilosort2',
            commit='43cbbfff89b9c88cdeb147ffd4ac35bfde9c7956'
        )

    def run(self):
        _keep_temp_files = True

        code = ''.join(random.choice(string.ascii_uppercase)
                       for x in range(10))
        tmpdir = os.environ.get('TEMPDIR', '/tmp') + '/kilosort2-tmp-' + code

        try:
            recording = SFMdaRecordingExtractor(self.recording_dir)
            if len(self.channels) > 0:
                recording = se.SubRecordingExtractor(
                    parent_recording=recording, channel_ids=self.channels)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)
            sorting = kilosort2_helper(
                recording=recording,
                tmpdir=tmpdir,
                detect_sign=self.detect_sign,
                adjacency_radius=self.adjacency_radius,
                detect_threshold=self.detect_threshold,
                merge_thresh=self.merge_thresh,
                freq_min=self.freq_min,
                freq_max=self.freq_max,
                pc_per_chan=self.pc_per_chan,
                minFR=self.minFR
            )
            SFMdaSortingExtractor.write_sorting(
                sorting=sorting, save_path=self.firings_out)
        except:
            if os.path.exists(tmpdir):
                if not _keep_temp_files:
                    print('removing tmpdir1')
                    shutil.rmtree(tmpdir)
            raise
        if not _keep_temp_files:
            print('removing tmpdir2')
            shutil.rmtree(tmpdir)


def kilosort2_helper(*,
                     recording,  # Recording object
                     tmpdir,  # Temporary working directory
                     detect_sign=-1,  # Polarity of the spikes, -1, 0, or 1
                     adjacency_radius=-1,  # Channel neighborhood adjacency radius corresponding to geom file
                     detect_threshold=6,  # Threshold for detection
                     merge_thresh=.98,  # Cluster merging threhold 0..1
                     freq_min=150,  # Lower frequency limit for band-pass filter
                     freq_max=6000,  # Upper frequency limit for band-pass filter
                     pc_per_chan=3,  # number of PC per chan
                     minFR=1 / 50
                     ):

    # # TODO: do not require ks2 to depend on irc -- rather, put all necessary .m code in the spikeforest repo
    # ironclust_path = os.environ.get('IRONCLUST_PATH_DEV', None)
    # if ironclust_path:
    #     print('Using ironclust from IRONCLUST_PATH_DEV directory: {}'.format(ironclust_path))
    # else:
    #     try:
    #         print('Auto-installing ironclust.')
    #         ironclust_path = install_ironclust(commit='042b600b014de13f6d11d3b4e50e849caafb4709')
    #     except:
    #         traceback.print_exc()
    #         raise Exception('Problem installing ironclust. You can set the IRONCLUST_PATH_DEV to force to use a particular path.')
    # print('For kilosort2, using ironclust utility functions from: {}'.format(ironclust_path))

    kilosort2_path = os.environ.get('KILOSORT2_PATH_DEV', None)
    if kilosort2_path:
        print('Using kilosort2 from KILOSORT2_PATH_DEV directory: {}'.format(kilosort2_path))
    else:
        try:
            print('Auto-installing kilosort2.')
            kilosort2_path = KiloSort2.install()
        except:
            traceback.print_exc()
            raise Exception('Problem installing kilosort2. You can set the KILOSORT2_PATH_DEV to force to use a particular path.')
    print('Using kilosort2 from: {}'.format(kilosort2_path))

    source_dir = os.path.dirname(os.path.realpath(__file__))

    dataset_dir = tmpdir + '/kilosort2_dataset'
    # Generate three files in the dataset directory: raw.mda, geom.csv, params.json
    SFMdaRecordingExtractor.write_recording(
        recording=recording, save_path=dataset_dir, _preserve_dtype=True)

    samplerate = recording.get_sampling_frequency()

    print('Reading timeseries header...')
    HH = mdaio.readmda_header(dataset_dir + '/raw.mda')
    num_channels = HH.dims[0]
    num_timepoints = HH.dims[1]
    duration_minutes = num_timepoints / samplerate / 60
    print('Num. channels = {}, Num. timepoints = {}, duration = {} minutes'.format(
        num_channels, num_timepoints, duration_minutes))

    print('Creating argfile.txt file...')
    txt = ''
    txt += 'samplerate={}\n'.format(samplerate)
    txt += 'detect_sign={}\n'.format(detect_sign)
    txt += 'adjacency_radius={}\n'.format(adjacency_radius)
    txt += 'detect_threshold={}\n'.format(detect_threshold)
    txt += 'merge_thresh={}\n'.format(merge_thresh)
    txt += 'freq_min={}\n'.format(freq_min)
    txt += 'freq_max={}\n'.format(freq_max)
    txt += 'pc_per_chan={}\n'.format(pc_per_chan)
    txt += 'minFR={}\n'.format(minFR)
    _write_text_file(dataset_dir + '/argfile.txt', txt)

    print('Running Kilosort2 in {tmpdir}...'.format(tmpdir=tmpdir))
    cmd = '''
        addpath('{source_dir}');
        addpath('{source_dir}/mdaio')
        try
            p_kilosort2('{ksort}', '{tmpdir}', '{raw}', '{geom}', '{firings}', '{arg}');
        catch
            quit(1);
        end
        quit(0);
        '''
    cmd = cmd.format(source_dir=source_dir, ksort=kilosort2_path,
                     tmpdir=tmpdir, raw=dataset_dir + '/raw.mda', geom=dataset_dir + '/geom.csv',
                     firings=tmpdir + '/firings.mda', arg=dataset_dir + '/argfile.txt')
    matlab_cmd = mlpr.ShellScript(cmd, script_path=tmpdir + '/run_kilosort2.m', keep_temp_files=True)
    matlab_cmd.write()
    shell_cmd = '''
        #!/bin/bash
        cd {tmpdir}
        echo '=====================' `date` '====================='
        matlab -nosplash -nodisplay -r run_kilosort2
    '''.format(tmpdir=tmpdir)
    shell_cmd = mlpr.ShellScript(shell_cmd, script_path=tmpdir + '/run_kilosort2.sh', keep_temp_files=True)
    shell_cmd.write(tmpdir + '/run_kilosort2.sh')
    shell_cmd.start()
    retcode = shell_cmd.wait()

    if retcode != 0:
        raise Exception('kilosort2 returned a non-zero exit code')

    # parse output
    result_fname = tmpdir + '/firings.mda'
    if not os.path.exists(result_fname):
        raise Exception('Result file does not exist: ' + result_fname)

    firings = mdaio.readmda(result_fname)
    sorting = se.NumpySortingExtractor()
    sorting.set_times_labels(firings[1, :], firings[2, :])
    return sorting


def _read_text_file(fname):
    with open(fname) as f:
        return f.read()


def _write_text_file(fname, str):
    with open(fname, 'w') as f:
        f.write(str)



import setuptools

pkg_name = "spikeforest"

# NOTE: you should install this project in development mode
# > python setup.py develop

setuptools.setup(
    name=pkg_name,
    version="0.9.0",
    author="Jeremy Magland",
    author_email="jmagland@flatironinstitute.org",
    description="Spike sorting",
    packages=setuptools.find_packages(),
    package_dir={
        'spikesorters': 'spikesorters',
        'forestview': 'forestview',
        'spikeforest_analysis': 'spikeforest_analysis',
        'spikeforest_common': 'spikeforest_common'
    },
    include_package_data=True,
    install_requires=[
        'numpy', 'scipy', 'matplotlib',
        'requests', 'pillow', 'pandas',
        'ipython', 'h5py', 'setuptools-git',
        'scikit-learn', 'python-frontmatter',
        'spikeextractors==0.4.2',
        'spiketoolkit==0.3.1'
    ],
    scripts=['bin/forestview'],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    )
)



import os
from mountaintools import client as mt
import mlprocessors as mlpr
import shutil


def install_ironclust(commit):
    spikeforest_alg_install_path = os.getenv('SPIKEFOREST_ALG_INSTALL_PATH', os.getenv('HOME') + '/spikeforest_algs')
    repo = 'https://github.com/jamesjun/ironclust'
    key = dict(
        alg='ironclust',
        repo=repo,
        commit=commit
    )
    source_path = spikeforest_alg_install_path + '/ironclust_' + commit
    if os.path.exists(source_path):
        hash0 = mt.computeDirHash(source_path)
        if hash0 == mt.getValue(key=key):
            print('IronClust is already auto-installed.')
            return source_path
        print('Removing directory: {}'.format(source_path))
        shutil.rmtree(source_path)

    script = """
    #!/bin/bash
    set -e

    git clone {repo} {source_path}
    cd {source_path}
    git checkout {commit}
    """.format(repo=repo, commit=commit, source_path=source_path)
    ss = mlpr.ShellScript(script=script)
    ss.start()
    retcode = ss.wait()
    if retcode != 0:
        raise Exception('Install script returned a non-zero exit code/')

    hash0 = mt.computeDirHash(source_path)
    mt.setValue(key=key, value=hash0)

    return source_path



from .bandpass_filter import bandpass_filter
import spikeextractors as se
import numpy as np


def estimateRecordingNoiseLevel(recording):
    return _estimate_noise_level(recording)


def autoScaleRecordingToNoiseLevel(recording, noise_level):
    noise_level_est = _estimate_noise_level(recording)
    return _scale_recording(recording, noise_level / noise_level_est)


def _estimate_channel_noise_levels_helper(recording):
    channel_ids = recording.get_channel_ids()
    # M=len(channel_ids)
    samplerate = int(recording.get_sampling_frequency())
    X = recording.get_traces(start_frame=samplerate * 1, end_frame=samplerate * 2)
    ret = []
    for ii in range(len(channel_ids)):
        # noise_level=np.std(X[ii,:])
        noise_level = np.median(np.abs(X[ii, :])) / 0.6745  # median absolute deviation (MAD)
        ret.append(noise_level)
    return ret


def _estimate_channel_noise_levels(recording):
    recording_filtered = bandpass_filter(recording=recording, freq_min=200, freq_max=30000)
    return _estimate_channel_noise_levels_helper(recording=recording_filtered)


def _estimate_noise_level(recording):
    return np.median(_estimate_channel_noise_levels(recording))


class ScaledRecording(se.RecordingExtractor):
    def __init__(self, recording, scalar=1):
        se.RecordingExtractor.__init__(self)
        if not isinstance(recording, se.RecordingExtractor):
            raise ValueError("'recording' must be a RecordingExtractor")
        self._recording = recording
        self._scalar = scalar
        self.copy_channel_properties(recording=self._recording)

    def get_sampling_frequency(self):
        return self._recording.get_sampling_frequency()

    def get_num_frames(self):
        return self._recording.get_num_frames()

    def get_channel_ids(self):
        return self._recording.get_channel_ids()

    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.get_num_frames()
        if channel_ids is None:
            channel_ids = self.get_channel_ids()
        traces = self._recording.get_traces(channel_ids=channel_ids, start_frame=start_frame, end_frame=end_frame)
        return traces * self._scalar


def _scale_recording(recording, scalar=1):
    return ScaledRecording(
        recording=recording, scalar=scalar
    )


def _compute_recording_range(recording):
    all_traces = recording.get_traces()
    return [np.min(all_traces), np.max(all_traces)]



from .filterrecording import FilterRecording
import numpy as np
from scipy import special


class BandpassFilterRecording(FilterRecording):
    def __init__(self, *, recording, freq_min, freq_max, freq_wid):
        FilterRecording.__init__(self, recording=recording, chunk_size=3000 * 10)
        self._recording = recording
        self._freq_min = freq_min
        self._freq_max = freq_max
        self._freq_wid = freq_wid
        self.copy_channel_properties(recording)

    def filterChunk(self, *, start_frame, end_frame):
        padding = 3000
        i1 = start_frame - padding
        i2 = end_frame + padding
        padded_chunk = self._read_chunk(i1, i2)
        if i1 < 0:
            for m in range(padded_chunk.shape[0]):
                padded_chunk[m, :-i1] = padded_chunk[m, -i1]
        if i2 > self._recording.get_num_frames():
            for m in range(padded_chunk.shape[0]):
                aa = (i2 - self._recording.get_num_frames())
                padded_chunk[m, -aa:] = padded_chunk[m, aa - 1]
        filtered_padded_chunk = self._do_filter(padded_chunk)
        return filtered_padded_chunk[:, start_frame - i1:end_frame - i1]

    def _create_filter_kernel(self, N, samplerate, freq_min, freq_max, freq_wid=1000):
        # Matches ahb's code /matlab/processors/ms_bandpass_filter.m
        # improved ahb, changing tanh to erf, correct -3dB pts  6/14/16
        T = N / samplerate  # total time
        df = 1 / T  # frequency grid
        relwid = 3.0  # relative bottom-end roll-off width param, kills low freqs by factor 1e-5.

        k_inds = np.arange(0, N)
        k_inds = np.where(k_inds <= (N + 1) / 2, k_inds, k_inds - N)

        fgrid = df * k_inds
        absf = np.abs(fgrid)

        val = np.ones(fgrid.shape)
        if freq_min != 0:
            val = val * (1 + special.erf(relwid * (absf - freq_min) / freq_min)) / 2  # pylint: disable=no-member
            val = np.where(np.abs(k_inds) < 0.1, 0, val)  # kill DC part exactly
        if freq_max != 0:
            val = val * (1 - special.erf((absf - freq_max) / freq_wid)) / 2  # pylint: disable=no-member
        val = np.sqrt(val)  # note sqrt of filter func to apply to spectral intensity not ampl
        return val

    def _do_filter(self, chunk):
        samplerate = self._recording.get_sampling_frequency()
        M = chunk.shape[0]
        chunk2 = chunk
        # Subtract off the mean of each channel unless we are doing only a low-pass filter
        # if self._freq_min!=0:
        #    for m in range(M):
        #        chunk2[m,:]=chunk2[m,:]-np.mean(chunk2[m,:])
        # Do the actual filtering with a DFT with real input
        chunk_fft = np.fft.rfft(chunk2)
        kernel = self._create_filter_kernel(
            chunk2.shape[1],
            samplerate,
            self._freq_min, self._freq_max, self._freq_wid
        )
        kernel = kernel[0:chunk_fft.shape[1]]  # because this is the DFT of real data
        chunk_fft = chunk_fft * np.tile(kernel, (M, 1))
        chunk_filtered = np.fft.irfft(chunk_fft)
        return chunk_filtered

    def _read_chunk(self, i1, i2):
        M = len(self._recording.get_channel_ids())
        N = self._recording.get_num_frames()
        if i1 < 0:
            i1b = 0
        else:
            i1b = i1
        if i2 > N:
            i2b = N
        else:
            i2b = i2
        ret = np.zeros((M, i2 - i1))
        ret[:, i1b - i1:i2b - i1] = self._recording.get_traces(start_frame=i1b, end_frame=i2b)
        return ret


def bandpass_filter(recording, freq_min=300, freq_max=6000, freq_wid=1000, resample=None):
    return BandpassFilterRecording(
        recording=recording,
        freq_min=freq_min,
        freq_max=freq_max,
        freq_wid=freq_wid,
    )



from copy import deepcopy
from mountaintools import client as mt
from spikeforest import SFMdaRecordingExtractor
from mountaintools import MountainClient
from .spikeforest_view_launchers import get_spikeforest_view_launchers
from .recordingcontext import RecordingContext

local_client = MountainClient()


class SpikeForestContext():
    def __init__(self, studies=[], recordings=[], sorting_results=[], aggregated_sorting_results=None):
        self._signal_handlers = dict()
        self._any_state_change_handlers = []

        print('******** FORESTVIEW: Initializing study context')
        self._studies = studies
        self._recordings = recordings
        self._sorting_results = sorting_results
        self._aggregated_sorting_results = aggregated_sorting_results
        self._recording_contexts = dict()

        self._studies_by_name = dict()
        for stu in self._studies:
            self._studies_by_name[stu['name']] = stu

        self._recordings_by_id = dict()
        for rec in self._recordings:
            id0 = rec['study'] + '/' + rec['name']
            self._recordings_by_id[id0] = rec
            c0 = RecordingContext(rec)
            c0.onAnyStateChanged(self._trigger_any_state_change_handlers)
            self._recording_contexts[id0] = c0

        for sr in self._sorting_results:
            rec = sr['recording']
            id0 = rec['study'] + '/' + rec['name']
            if id0 in self._recording_contexts:
                rc = self._recording_contexts[id0]
                rc.addSortingResult(sr)

        print('******** FORESTVIEW: Done initializing study context')
        self._state = dict(
            current_recording_id=None,
            selected_recording_ids=[]
        )

    def viewLaunchers(self):
        return get_spikeforest_view_launchers(self)

    def studyObject(self, name):
        return deepcopy(self._studies_by_name.get(name, {}))

    def studySetNameForStudy(self, name):
        return self.studyObject(name).get('study_set')

    def studyNames(self):
        return sorted(list())

    def recordingIds(self):
        return sorted(list(self._recordings_by_id.keys()))

    def recordingObject(self, recid):
        return deepcopy(self._recordings_by_id[recid])

    def recordingContext(self, recid):
        if not recid:
            return None
        return self._recording_contexts[recid]

    def hasAggregatedSortingResults(self):
        return self._aggregated_sorting_results is not None

    def aggregatedSortingResults(self):
        if type(self._aggregated_sorting_results) == str:
            obj = mt.loadObject(path=self._aggregated_sorting_results)
            return obj
        else:
            return self._aggregated_sorting_results

    # def recordingExtractor(self, recording_name, *, download):
    #     print('loading recording extractor....', recording_name, download)
    #     return SFMdaRecordingExtractor(self._study_dir + '/' + recording_name, download=download)

    def onAnyStateChanged(self, handler):
        for key in self._state.keys():
            self._register_state_change_handler(key, handler)
        self._any_state_change_handlers.append(handler)

    def stateObject(self):
        ret = deepcopy(self._state)
        rc = self.recordingContext(self.currentRecordingId())
        if rc:
            ret['current_recording'] = rc.stateObject()
        else:
            ret['current_recording'] = None
        return ret

    # current recording
    def currentRecordingId(self):
        return self._get_state_value('current_recording_id')

    def setCurrentRecordingId(self, recname):
        self._set_state_value('current_recording_id', recname)

    def onCurrentRecordingChanged(self, handler):
        self._register_state_change_handler('current_recording_id', handler)

    # selected recordings
    def setSelectedRecordingIds(self, recids):
        if recids is None:
            recids = []
        recids = sorted(recids)
        self._set_state_value('selected_recording_ids', recids)

    def selectedRecordingIds(self):
        return self._get_state_value('selected_recording_ids')

    def onSelectedRecordingsChanged(self, handler):
        self._register_state_change_handler('selected_recording_ids', handler)

    def _set_state_value(self, name, val):
        if self._state[name] == val:
            return
        self._state[name] = deepcopy(val)
        self._emit('state-changed-' + name)

    def _get_state_value(self, name):
        return deepcopy(self._state[name])

    def _register_state_change_handler(self, name, handler):
        self._register_signal_handler('state-changed-' + name, handler)

    def _register_signal_handler(self, signal_name, handler):
        if signal_name not in self._signal_handlers:
            self._signal_handlers[signal_name] = []
        self._signal_handlers[signal_name].append(handler)

    def _emit(self, signal_name):
        if signal_name in self._signal_handlers:
            for handler in self._signal_handlers[signal_name]:
                handler()

    def _trigger_any_state_change_handlers(self):
        for handler in self._any_state_change_handlers:
            handler()



import vdomr as vd


class AnalysisSummaryView(vd.Component):
    def __init__(self, context, opts=None, prepare_result=None):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)

    @staticmethod
    def prepareView(context, opts):
        context.initialize()

    def setSize(self, size):
        self._size = size

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Analysis summary'

    def render(self):
        # row info for the table
        rr = []
        rr.append(dict(label='Recording groups', value=', '.join(self._context.recordingGroups())))
        rr.append(dict(label='Sorters to run', value=', '.join(self._context.sorterKeys())))
        rr.append(dict(label='Output path', value=self._context.outputPath()))
        rr.append(dict(label='Download from', value=', '.join(self._context.downloadFrom())))
        rr.append(dict(label='Job timeout (sec)', value=self._context.jobTimeout()))
        rr.append(dict(label='Compute resources', value=', '.join(self._context.computeResourceKeys())))
        rr.append(dict(label='Sorter definition keys', value=', '.join(self._context.sorterDefinitionKeys())))

        rows = []
        for r in rr:
            rows.append(vd.tr(
                vd.th(r['label']), vd.td(r['value'])
            ))

        table = vd.table(rows, style={
                         'text-align': 'left', 'width': 'auto', 'font-size': '13px'}, class_='table')

        return vd.components.ScrollArea(vd.div(
            vd.h2('Analysis: {}'.format(self._context.analysisName())),
            table, height=self._size[1]
        ))



from matplotlib import pyplot as plt
from PIL import Image
import os

def savePlot(fname,quality=40,close_figure=True):
    plt.savefig(fname+'.png')
    if close_figure:
      plt.close()
    im=Image.open(fname+'.png').convert('RGB')
    os.remove(fname+'.png')
    im.save(fname,quality=quality)